# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides standard wrappings for ML models.
"""
import importlib
from copy import deepcopy
from typing import Any
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np

from deel.puncc.api import nonconformity_scores
from deel.puncc.api.utils import dual_predictor_check
from deel.puncc.api.utils import supported_types_check

if importlib.util.find_spec("tensorflow") is not None:
    import tensorflow as tf


class BasePredictor:
    """Wrapper of a point prediction model :math:`\hat{f}`. Enables to
    standardize the interface of predictors and to expose generic :func:`fit`,
    :func:`predict` and :func:`copy` methods.

    :param Any model: prediction model :math:`\hat{f}`
    :param bool is_trained: boolean flag that informs if the model is
        pre-trained. If True, the call to :func:`fit` will be skipped
    :param compile_kwargs: keyword arguments to be used if needed during the
        call :func:`model.compile` on the underlying model

    .. _example basepredictor:

    Sklearn regression examples::

        from deel.puncc.api.prediction import BasePredictor
        from sklearn.ensemble import RandomForestRegressor
        import numpy as np

        # Generate data
        X_train = np.random.uniform(0, 10, 1000)
        X_new = np.random.uniform(0, 10, 100)
        y_train = np.sin(X)

        # Instantiate two random forest models composed of 100 trees.
        rf_model1 = RandomForestRegressor(n_estimators=100)
        rf_model2 = RandomForestRegressor(n_estimators=100)

        # Consider that `rf_model2` is previously trained
        rf_model2.fit(X_train, y_train)

        # We will instantiate two wrappers:
        # - `predictor1` will wrap `rf_model1`.
        # - `predictor2` will wrap `rf_model2`. Also, we don't want to retrain our model,
        #   so we will specify it to the constructor by setting the argument
        #   `is_trained` to True. If `fit` is called, it will be skipped.
        # The argument `is_trained` defaults to False.
        predictor1 = BasePredictor(rf_model1, is_trained=False)
        predictor2 = BasePredictor(rf_model2, is_trained=True)

        # Fit `predictor2` to the training data.
        # No need to call fit on `predictor2`. But if you do, it will be skipped.
        predictor1.fit(X_train,y_train)

        # Predict on X_new
        y_pred1 = predictor1.predict(X_new)
        y_pred2 = predictor2.fit(X_new)


    Keras classification example::

        from deel.puncc.api.prediction import BasePredictor
        import tensorflow as tf
        import numpy as np

        # Generate data
        X_train = np.random.uniform(0, 10, 1000)
        X_train = np.expand_dims(X_train,-1)
        X_new = np.random.uniform(0, 10, 100)
        y_train = np.array([1 if x>5 else 0 for x in X_train])
        y_train = tf.keras.utils.to_categorical(y_train)

        # Instantiate a classifier as sequential model and add two dense layers
        cl_model = tf.keras.Sequential()
        cl_model.add(tf.keras.layers.Dense(10, activation="relu"))
        cl_model.add(tf.keras.layers.Dense(2, activation="softmax"))

        # The compile options need be passed to the constructor `BasePredictor`.
        # The wrapper will call compile(**compile_kwargs) on an internal copy of the model.
        # Our model is a classifier, we use categorical crossentropy as loss function.
        compile_kwargs={"optimizer":"rmsprop", "loss":"categorical_crossentropy"}
        predictor = BasePredictor(cl_model, is_trained=False, **compile_kwargs)

        # The fit method is provided with a given training dataset (X,y) and
        # with the train configuration of the underlying model. In the example below,
        # we train the model over 5 epochs on batches containing 128 examples.
        predictor.fit(X_train, y_train, **{"epochs":5,"batch_size":128})

        # The `BasePredictor.predict` method enables to pass keyword arguments
        # to the `predict`call of the underlying model (e.g., verbose).
        y_pred = predictor.predict(X_new, **{"verbose":1})

    """

    def __init__(self, model: Any, is_trained: bool = False, **compile_kwargs):
        self.model = model
        self.is_trained = is_trained
        self.compile_kwargs = compile_kwargs

        if self.compile_kwargs:
            _ = self.model.compile(**self.compile_kwargs)

    def get_is_trained(self) -> bool:
        """Get flag that informs if the model is pre-trained."""
        return self.is_trained

    def fit(self, X: Iterable, y: Optional[Iterable] = None, **kwargs) -> None:
        """Fit model to the training data.

        :param Iterable X: train features.
        :param Optional[Iterable] y: train labels. Defaults to None (unsupervised).
        :param kwargs: keyword arguments to be passed to the call :func:`fit`
            on the underlying model :math:`\hat{f}`.

        .. note::

            For more details, check this :ref:`code snippets
            <example basepredictor>`.

        """
        if y is None:
            self.model.fit(X, **kwargs)
        else:
            self.model.fit(X, y, **kwargs)
        # self.is_trained = True

    def predict(self, X: Iterable, **kwargs) -> np.ndarray:
        """Compute predictions on new examples.

        :param Iterable X: new examples' features.
        :param dict kwargs: predict configuration to be passed to the `predict`
            method of the underlying model :math:`\hat{f}`.

        :returns: predictions :math:`\hat{f}(X)` associated to the new
            examples X.
        :rtype: ndarray

        .. note::

            For more details, check this :ref:`code snippets
            <example basepredictor>`.

        """
        # Remove axis of length one to avoid broadcast in computation
        # of non-conformity scores
        return np.squeeze(self.model.predict(X, **kwargs))

    def copy(self):
        """Returns a copy of the predictor. The underlying model is either
        cloned (Keras model) or deepcopied (sklearn and similar models).

        :returns: copy of the predictor.
        :rtype: BasePredictor

        :raises RuntimeError: copy unsupported for provided models.

        """

        model_type_str = str(type(self.model))

        if (
            "tensorflow" in model_type_str
            or "keras" in model_type_str
            and importlib.util.find_spec("tensorflow") is not None
        ):
            # pylint: disable=E1101
            model = tf.keras.models.clone_model(self.model)
            try:
                model.set_weights(self.model.get_weights())
            except Exception:
                pass
        else:
            model = deepcopy(self.model)

        predictor_copy = self.__class__(
            model=model, is_trained=self.is_trained, **self.compile_kwargs
        )
        return predictor_copy


class IdPredictor(BasePredictor):
    """Subclass of :class:`BasePredictor` to directly wrap existing predictions.
    The predictions are directly returned without any modification.

    :param model: model to be wrapped.

    .. _example idpredictor:

    Conformal regression example::

        import numpy as np
        from deel.puncc.api.prediction import IdPredictor
        from deel.puncc.regression import SplitCP
        from deel.puncc.metrics import regression_mean_coverage, regression_sharpness
        from deel.puncc.plotting import plot_prediction_intervals

        # Generate data
        X = np.linspace(0, 20, 5000)
        # randomly shuffle the data
        np.random.shuffle(X)
        X_calib, X_test = X[:4000], X[1000:]


        # Define the real target function
        def real_f(X):
            return 2 * X + np.random.randn(len(X)) * X * 0.5


        # Target values for the calibration and new data
        y_calib = real_f(X_calib)
        y_test = real_f(X_test)


        # Suppose we can obtain predictions from an API call
        def api_call(X):
            return 2 * X  # This is a remote model that estimates of the target


        # The model can still be conformalized using the calibration data as follows
        ## 1. Instantiate the Predictor wrapper, which serves as a container to host the predictions
        dummy_predictor = IdPredictor()

        ## 2. CP method initialization
        split_cp = SplitCP(
            dummy_predictor, train=False
        )  # train=False to avoid trying to retrain the model internally

        ## 3. Request predictions on the calibration set
        y_pred = api_call(X_calib)

        ## 4. The call to fit computes the nonconformity scores on the
        ## calibration set. Instead of features, we need to provide the
        ## predictions as X_calib along with the true target values.
        split_cp.fit(X_calib=y_pred, y_calib=y_calib)

        ## 5. The predict method infers prediction intervals with respect to
        ## the significance level alpha = 10%. Make sure you provide the
        ## point predictions on the test set.
        y_pred, y_pred_lower, y_pred_upper = split_cp.predict(
            api_call(X_test), alpha=0.1
        )

        # Compute marginal coverage and average width of the prediction intervals
        coverage = regression_mean_coverage(y_test, y_pred_lower, y_pred_upper)
        width = regression_sharpness(
            y_pred_lower=y_pred_lower, y_pred_upper=y_pred_upper
        )
        print(f"Marginal coverage: {np.round(coverage, 2)}")
        print(f"Average width: {np.round(width, 2)}")

        ax = plot_prediction_intervals(
            y_test, y_pred_lower, y_pred_upper, X=X_test, y_pred=y_pred
        )





    """

    def __init__(self, model=None, **kwargs):
        self.kwargs = kwargs
        super().__init__(model, is_trained=True)

    def fit(self, X: Iterable, y: Optional[Iterable] = None, **kwargs) -> None:
        """Fit method is not supported for IdPredictor. Raises a RuntimeError."""
        raise RuntimeError(
            "IdPredictor does not support fit method. Use BasePredictor in"
            + " case you need to train the underlying model."
        )

    def predict(self, X: Iterable):
        """Returns the input argument as output data.

        :param Iterable X: predictions.

        :return: predictions.
        :rtype: np.ndarray
        """
        return X

    def predict_with_model(self, X):
        """
        Predicts the output using the wrapped model.

        :param Iterable X: the input features to build predictions.

        :return: predictions.
        :rtype: np.ndarray
        """
        return self.model.predict(X)


class DualPredictor:
    """Wrapper of **two** joint point prediction models
    :math:`(\hat{f_1},\hat{f_2})`.
    The prediction :math:`\hat{y}` of a :class:`DualPredictor` is a tuple
    :math:`\hat{y}=(\hat{y}_1,\hat{y}_2)`, where :math:`\hat{y}_1`
    (resp. :math:`\hat{y}_2`) is the prediction of :math:`\hat{f}_1`
    (resp. :math:`\hat{f}_2`).

    Enables to standardize the interface of predictors and to expose generic
    :func:`fit`, :func:`predict` and :func:`copy` methods.

    :param List[Any] model: list of two prediction models
        :math:`[\hat{f_1},\hat{f_2}]`.
    :param List[bool] is_trained: list of boolean flag that informs if the
        models are pre-trained. True value will skip the fitting of the
        corresponding model.
    :param List compile_kwargs: list of keyword arguments to be used if needed
        to compile the underlying models.

    .. _example dualpredictor:

    Joint conditional mean / quantile regression example::

        from deel.puncc.api.prediction import DualPredictor
        import tensorflow_addons as tfa
        import tensorflow as tf
        from sklearn.ensemble import RandomForestRegressor

        # Instantiate conditional mean model
        rf_model = RandomForestRegressor(n_estimators=100)

        # Instantiate 90-th quantile model
        q_model = tf.keras.Sequential()
        q_model.add(tf.keras.layers.Dense(10, activation="relu"))
        q_model.add(tf.keras.layers.Dense(1, activation="relu"))
        pinball_loss=tfa.losses.PinballLoss(tau=.9)
        compile_kwargs={'optimizer':'sgd', 'loss':pinball_loss}

        # The compile options need be passed to the constructor `DualPredictor`.
        # The wrapper will call compile on the internal copy of each model if needed.
        # Our predictor combines two regressors, of which only the second needs to be compiled.
        tf_predictor = DualPredictor(models=[rf_model, q_model],
                                     compile_args=[{}, compile_kwargs])

        # The fit method is provided with a given training dataset (X,y) and
        # the train configurations of the underlying models. In the example below,
        # no specific configuration is passed to `rf_model` whereas we want to
        # train the `q_model` over 10 epochs on batches containing 32 examples.
        predictor.fit(X_train, y_train, [{}, {"epochs":10,"batch_size":32}])

        # The `DualPredictor.predict` method enables to pass keyword arguments
        # to the `predict` call of the underlying models. In the example below,
        # we want to turn off the verbosity of `q_model.predict`.
        # Besides, `y_pred` consists of a couple (y1, y2) for each new example.
        # If `X_new` is a (n,m) matrix, the shape of `y_pred` will be (n, 2).
        y_pred = predictor.predict(X_new, [{}, {"verbose":0}])


    """

    def __init__(
        self,
        models: List[Any],
        is_trained: List[bool] = [False, False],
        compile_args: List[dict] = [{}, {}],
    ):
        dual_predictor_check(models, "models", "models")
        dual_predictor_check(is_trained, "is_trained", "booleans")
        dual_predictor_check(compile_args, "compile_args", "dictionnaries")
        self.models = models
        self.is_trained = is_trained
        self.compile_args = compile_args

        if len(self.compile_args[0].keys()) != 0 and is_trained[0] is False:
            _ = self.models[0].compile(**self.compile_args[0])

        if len(self.compile_args[1].keys()) != 0 and is_trained[1] is False:
            _ = self.models[1].compile(**self.compile_args[1])

    def get_is_trained(self) -> bool:
        """Get flag that informs if the models are pre-trained.
        Returns True only when both models are pretrained.
        """
        return self.is_trained[0] and self.is_trained[1]

    def fit(
        self, X: Iterable, y: Iterable, dictargs: List[dict] = [{}, {}]
    ) -> None:
        """Fit model to the training data.

        :param Iterable X: train features.
        :param Iterable y: train labels.
        :param List[dict[str]] dictargs: list of fit configurations to be
            passed to the `fit` method of the underlying models
            :math:`\hat{f}_1` and :math:`\hat{f}_2`, respectively.

        .. note::

            For more details, check this :ref:`code snippet
            <example dualpredictor>`.

        """
        dual_predictor_check(dictargs, "dictargs", "dictionnaries")
        for count, model in enumerate(self.models):
            if not self.is_trained[count]:
                model.fit(X, y, **dictargs[count])

    def predict(self, X, dictargs: List[dict] = [{}, {}]) -> Tuple[np.ndarray]:
        """Compute predictions on new examples.

        :param Iterable X: new examples' features.
        :param dict kwargs: list of predict configurations to be passed to the
            `predict` method of the underlying models :math:`\hat{f}_1` and
            :math:`\hat{f}_2`, respectively.

        :returns: predictions :math:`\hat{y}=\hat{f}(X)` associated to the new
            examples X. For an instance :math:`X_i`, the prediction consists of
            a couple :math:`\hat{f}(X_i)=(\hat{f}_1(X_i), \hat{f}_2(X_i))`.
        :rtype: Tuple[ndarray]

        :raises NotImplementedError: predicted values not formated as numpy
            ndarrays.

        .. note::

            For more details, check this :ref:`code snippet
            <example dualpredictor>`.

        """
        dual_predictor_check(dictargs, "dictargs", "dictionnaries")
        model1_pred = self.models[0].predict(X, **dictargs[0])
        model2_pred = self.models[1].predict(X, **dictargs[1])
        supported_types_check(model1_pred, model2_pred)

        if isinstance(model1_pred, np.ndarray):
            Y_pred = np.column_stack((model1_pred, model2_pred))
        else:  # In case the models return something other than an np.ndarray
            raise NotImplementedError(
                "Predicted values must be of type numpy.ndarray."
            )

        return np.squeeze(Y_pred)

    def copy(self):
        """Returns a copy of the predictor. The underlying models are either
        cloned (Keras model) or deepcopied (sklearn and similar models).

        :returns: copy of the predictor.
        :rtype: DualPredictor

        :raises RuntimeError: copy unsupported for provided models.

        """
        models_copy = []
        for model in self.models:
            try:
                model_copy = deepcopy(model)
            except Exception as e_outer:
                if importlib.util.find_spec("tensorflow") is not None:
                    try:
                        # pylint: disable=E1101
                        model_copy = tf.keras.models.clone_model(model)
                    except Exception as e_inner:
                        msg = (
                            f"Cannot copy models. Many possible reasons:\n"
                            f" 1- {e_inner} \n 2- {e_outer}"
                        )
                        raise RuntimeError(msg)
                else:
                    raise Exception(e_outer)
            models_copy.append(model_copy)

        predictor_copy = self.__class__(
            models=models_copy,
            is_trained=self.is_trained,
            compile_args=self.compile_args,
        )
        return predictor_copy


class MeanVarPredictor(DualPredictor):
    """Subclass of :class:`DualPredictor` to specifically wrap a conditional
    mean estimator :math:`\hat{\mu}` and a conditional dispersion estimator
    :math:`\hat{\sigma}`.\n

     Specifically, the dispersion model :math:`\hat{\sigma}` is trained on the
     mean absolute deviation of :math:`\hat{\mu}`'s predictions from the true
     labels :math:`y`. Given two training algorithms :math:`{\cal A}_1` and
     :math:`{\cal A}_2` and a training dataset :math:`(X_{train}, y_{train})`:

         .. math::

             \hat{\mu} \Leftarrow {\cal A}_1(X_{train}, y_{train})

         .. math::
             \hat{\sigma} \Leftarrow {\cal A}_2(X_{train},
             |\hat{\mu}(X_{train})-y_{train}|)


     :param List[Any] model: list of two prediction models
        :math:`[\hat{\mu},\hat{\sigma}]`.
     :param List[bool] is_trained: list of boolean flag that informs if the
        models are pre-trained. True value will skip the fitting of the
        corresponding model.
     :param List compile_kwargs: list of keyword arguments to be used if
        needed to compile the underlying models :math:`\hat{\mu}` and
        :math:`\hat{\sigma}`, respectively.

     .. _example MeanVarPredictor:

    Here follows an example of wrapping conditional mean and dispersion models::

        from deel.puncc.api.prediction import MeanVarPredictor
        from sklearn.ensemble import RandomForestRegressor

        # Instantiate conditional mean model
        mu_model = linear_model.LinearRegression()
        # Instantiate conditional dispersion model
        sigma_model = RandomForestRegressor(
            n_estimators=100, random_state=random_seed
        )

        # The instantiation of a :class:`MeanVarPredictor` is simple as the
        # selected models do not require any compilation
        mean_var_predictor = MeanVarPredictor([mu_model, sigma_model])

        # The fit method is provided with a given training dataset (X,y).
        # We do not choose any specific the train configurations.
        mean_var_predictor.fit(X_train, y_train)

        # The method `predict` yields `y_pred` that consists of a couple
        # (y1, y2) for each new example.
        # If `X_new` is a (n,m) matrix, the shape of `y_pred` will be (n, 2).
        y_pred = mean_var_predictor.predict(X_new)


    To see an example how to pass compilation/fit/predict configurations as
    arguments to the underlying models, check this
    :ref:`code snippet <example DualPredictor>`.

    """

    def fit(
        self, X: Iterable, y: Iterable, dictargs: List[dict] = [{}, {}]
    ) -> None:
        """Fit models to the training data. The dispersion model
        :math:`\hat{\sigma}` is trained on the mean absolute deviation of
        :math:`\hat{mu}`'s predictions :math:`\hat{\mu}` from the true labels
        :math:`y`.

        :param Iterable X: train features.
        :param Iterable y: train labels.
        :param List[dict] dictargs: list of fit configurations to be passed to
            the `fit` method of the underlying models :math:`\hat{\mu}` and
            :math:`\hat{\sigma}`, respectively.
        """
        self.models[0].fit(X, y, **dictargs[0])
        mu_pred = self.models[0].predict(X)
        mads = nonconformity_scores.absolute_difference(mu_pred, y)
        self.models[1].fit(X, mads, **dictargs[1])



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the core calibrator, providing a structure to estimate
the nonconformity scores on the calibration set and to compute the prediction
sets.
"""
import logging
import warnings
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Tuple

import numpy as np

from deel.puncc.api.corrections import bonferroni
from deel.puncc.api.utils import alpha_calib_check
from deel.puncc.api.utils import quantile

logger = logging.getLogger(__name__)


class BaseCalibrator:
    """:class:`BaseCalibrator` offers a framework to compute user-defined
    nonconformity scores on calibration dataset(s) (:func:`fit`) and to use for
    constructing and/or calibrating prediction sets (:func:`calibrate`).

    :param callable nonconf_score_func: nonconformity score function.
    :param callable pred_set_func: prediction set construction function.
    :param callable weight_func: function that takes as argument an array of
        features X and returns associated "conformality" weights,
        defaults to None.

    :raises NotImplementedError: provided :data:`weight_method` is not suitable.

    .. _example basecalibrator:

    **Regression calibrator example:**


    Consider a pretrained model :math:`\hat{f}`, a calibration dataset
    :math:`(X_{calib}, y_{calib})` and a test dataset :math:`(X_{test}, y_{test})`.
    The model :math:`\hat{f}` generates predictions on the calibration and
    test sets:

    .. math::
        y_{pred, calib}=\hat{f}(X_{calib})

    .. math::
        y_{pred, test}=\hat{f}(X_{test})

    Two function need to be defined before instantiating the
    :class:`BaseCalibrator`: a nonconformity score function and a definition of
    how the prediction sets are computed. In the example below, these are
    implemented from scratch but a collection of ready-to-use nonconformity
    scores and prediction sets are provided in the modules
    :ref:`nonconformity_scores <nonconformity_scores>` and
    :ref:`prediction_sets <prediction_sets>`, respectively.


    .. code-block:: python

        from deel.puncc.api.calibration import BaseCalibrator
        import numpy as np

        # First, we define a nonconformity score function that takes as argument
        # the predicted values y_pred = model(X) and the true labels y_true. In
        # this example, we reimplement the mean absolute deviation that is
        # already defined in `deel.puncc.api.nonconformity_scores.mad`
        def nonconformity_function(y_pred, y_true):
            return np.abs(y_pred - y_true)

        # Prediction sets are computed based on points predictions and
        # the quantiles of nonconformity scores. The function below returns a
        # fixed size interval around the point predictions.
        def prediction_set_function(y_pred, scores_quantile):
            y_lo = y_pred - scores_quantile
            y_hi = y_pred + scores_quantile
            return y_lo, y_hi

        # The calibrator is instantiated by passing the two functions defined
        # above to the constructor.
        calibrator = BaseCalibrator(
            nonconf_score_func=nonconformity_function,
            pred_set_func=prediction_set_function
        )

        # Generate dummy data and predictions
        y_pred_calib = np.random.rand(1000)
        y_true_calib = np.random.rand(1000)
        y_pred_test = np.random.rand(1000)

        # The nonconformity scores are computed by calling the `fit` method
        # on the calibration dataset.
        calibrator.fit(y_pred=y_pred_calib, y_true=y_true_calib)

        # The lower and upper bounds of the prediction interval are then returned
        # by the call to calibrate on the new data w.r.t a risk level of 10%.
        y_pred_low, y_pred_high = calibrator.calibrate(y_pred=y_pred_test, alpha=.1)


    """

    def __init__(
        self,
        *,
        nonconf_score_func: Callable,
        pred_set_func: Callable,
        weight_func: Callable = None,
    ):
        self.nonconf_score_func = nonconf_score_func
        self.pred_set_func = pred_set_func
        self.weight_func = weight_func
        self._len_calib = 0
        self._residuals = None
        self._norm_weights = None
        self._feature_axis = None

    def fit(
        self,
        *,
        y_true: Iterable,
        y_pred: Iterable,
    ) -> None:
        """Compute and store nonconformity scores on the calibration set.

        :param Iterable y_true: true labels.
        :param Iterable y_pred: predicted values.

        """
        # TODO check structure match in supported types
        logger.info("Computing nonconformity scores ...")
        self._residuals = self.nonconf_score_func(y_pred, y_true)
        self._len_calib = len(self._residuals)
        if (
            y_pred.ndim > 1
        ):  # TODO make sure that ndim can be applied to all types
            self._feature_axis = -1
        logger.debug("Nonconformity scores computed !")

    def calibrate(
        self,
        *,
        alpha: float,
        y_pred: Iterable,
        weights: Optional[Iterable] = None,
        correction: Optional[Callable] = bonferroni,
    ) -> Tuple[np.ndarray]:
        """Compute calibrated prediction sets for new examples w.r.t a
        significance level :math:`\\alpha`.

        :param float alpha: significance level (max miscoverage target).
        :param Iterable y_pred: predicted values.
        :param Iterable weights: weights to be associated to the nonconformity
                                 scores. Defaults to None when all the scores
                                 are equiprobable.
        :param Callable correction: correction for multiple hypothesis testing
                                    in the case of multivariate regression.
                                    Defaults to Bonferroni correction.

        :returns: prediction set.
                  In case of regression, returns (y_lower, y_upper).
                  In case of classification, returns (classes,).
        :rtype: Tuple[ndarray]

        :raises RuntimeError: :meth:`calibrate` called before :meth:`fit`.
        :raise ValueError: failed check on :data:`alpha` w.r.t size of the
            calibration set.

        """
        residuals_Q = self.compute_quantile(
            alpha=alpha, weights=weights, correction=correction
        )

        return self.pred_set_func(y_pred, scores_quantile=residuals_Q)

    def set_norm_weights(self, norm_weights: np.ndarray) -> None:
        """Setter of normalized weights associated to the nonconformity
        scores on the calibration set.

        :param ndarray norm_weights: normalized weights array

        """
        self._norm_weights = norm_weights

    def get_norm_weights(self) -> np.ndarray:
        """Getter of normalized weights associated to the nonconformity
        scores on the calibration set.

        :returns: normalized weights of nonconformity scores.
        :rtype: np.ndarray
        """
        return self._norm_weights

    def get_nonconformity_scores(self) -> np.ndarray:
        """Getter of computed nonconformity scores on the calibration set.

        :returns: nonconformity scores.
        :rtype: np.ndarray
        """
        return self._residuals

    def compute_quantile(
        self,
        *,
        alpha: float,
        weights: Optional[Iterable] = None,
        correction: Optional[Callable] = bonferroni,
    ) -> np.ndarray:
        """Compute quantile of scores w.r.t a
        significance level :math:`\\alpha`.

        :param float alpha: significance level (max miscoverage target).
        :param Iterable weights: weights to be associated to the nonconformity
                                 scores. Defaults to None when all the scores
                                 are equiprobable.
        :param Callable correction: correction for multiple hypothesis testing
                                    in the case of multivariate regression.
                                    Defaults to Bonferroni correction.

        :returns: quantile
        :rtype: ndarray

        :raises RuntimeError: :meth:`compute_quantile` called before :meth:`fit`.
        :raise ValueError: failed check on :data:`alpha` w.r.t size of the
            calibration set.

        """

        if self._residuals is None:
            raise RuntimeError("Run `fit` method before calling `calibrate`.")

        alpha = correction(alpha)

        # Check consistency of alpha w.r.t the size of calibration data
        if weights is None:
            alpha_calib_check(alpha=alpha, n=self._len_calib)

        # Compute weighted quantiles
        ## Lemma 1 of Tibshirani's paper (https://arxiv.org/pdf/1904.06019.pdf)
        ## The coverage guarantee holds with 1) the inflated
        ## (1-\alpha)(1+1/n)-th quantile or 2) when adding an infinite term to
        ## the sequence and computing the $(1-\alpha)$-th empirical quantile.
        if self._residuals.ndim > 1:
            infty_array = np.full((1, self._residuals.shape[-1]), np.inf)
        else:
            infty_array = np.array([np.inf])
        lemma_residuals = np.concatenate((self._residuals, infty_array), axis=0)
        residuals_Q = quantile(
            lemma_residuals,
            1 - alpha,
            w=weights,
            feature_axis=self._feature_axis,
        )

        return residuals_Q

    @staticmethod
    def barber_weights(weights: np.ndarray) -> np.ndarray:
        """Compute and normalize inference weights of the nonconformity distribution
        based on `Barber et al. <https://arxiv.org/abs/2202.13415>`_.

        :param ndarray weights: weights assigned to the samples.

        :returns: normalized weights.
        :rtype: ndarray
        """

        weights_len = len(weights)

        # Computation of normalized weights
        sum_weights = np.sum(weights)
        w_norm = np.zeros(weights_len + 1)
        w_norm[:weights_len] = weights / (sum_weights + 1)
        w_norm[weights_len] = 1 / (sum_weights + 1)

        return w_norm


class ScoreCalibrator:
    """:class:`ScoreCalibrator` offers a framework to compute user-defined
    scores on a calibration dataset (:func:`fit`) and to test the conformity
    of new data points (:func:`is_conformal`) with respect to a significance
    (error) level :math:`\\alpha`. Such calibrator can be used for example
    to calibrate the decision threshold of anomaly detection scores.

    :param callable nonconf_score_func: nonconformity score function.
    :param callable weight_func: function that takes as argument an array of
        data points and returns associated "conformality" weights,
        defaults to None.

    .. _example scorecalibrator:

    **Anomaly detection example:**


    Consider the two moons dataset. We want to detect anomalous points in a new
    sample generated following a uniform distribution. The LOF algorithm is
    used to obtain anomaly scores; then a :class:`ScoreCalibrator` is
    instantiated to decide which scores are conformal (not anomalies) with
    respect to a significance level :math:`\\alpha`.

    .. code-block:: python

        import numpy as np
        from sklearn.datasets import make_moons
        from sklearn.model_selection import train_test_split
        from sklearn.neighbors import LocalOutlierFactor
        import matplotlib.pyplot as plt

        from deel.puncc.api.calibration import ScoreCalibrator

        # First, we generate the two moons dataset
        dataset = 4*make_moons(n_samples=1000, noise=0.05,
            random_state=0)[0] - np.array([0.5, 0.25])

        # Split data into proper fitting and calibration sets
        fit_set, calib_set = train_test_split(dataset, train_size=.8)

        # Generate new data points
        rng = np.random.RandomState(42)
        new_samples = rng.uniform(low=-6, high=8, size=(200, 2))

        # Instantiate the LOF anomaly detection algorithm
        algorithm = LocalOutlierFactor(n_neighbors=35, novelty=True)

        # Fit the LOF on the proper fitting dataset
        algorithm.fit(X=fit_set)

        # The nonconformity scores are defined as the LOF (anomaly) scores.
        # By default, score_samples return the opposite of LOF scores.
        ncf = lambda X: -algorithm.score_samples(X)

        # The ScoreCalibrator is instantiated by passing the LOF score function
        # to the constructor
        cad = ScoreCalibrator(nonconf_score_func=ncf)

        # The LOF scores are computed by calling the `fit` method
        # on the calibration dataset
        cad.fit(calib_set)

        # We set the target false detection rate to 1%
        alpha = .01

        # The method `is_conformal` is called on the new data points
        # to test which are conformal (not anomalous) and which are not
        results = cad.is_conformal(z=new_samples, alpha=alpha)
        not_anomalies = new_samples[results]
        anomalies = new_samples[np.invert(results)]

        # Plot the results
        plt.scatter(calib_set[:,0], calib_set[:,1],
                    s=10, label="Inliers")
        plt.scatter(not_anomalies[:, 0], not_anomalies[:, 1], s=40, marker="x",
                    color="blue", label="Normal")
        plt.scatter(anomalies[:, 0], anomalies[:, 1], s=40, marker="x",
                    color="red", label="Anomaly")
        plt.xticks(())
        plt.yticks(())
        plt.legend(loc="lower left")

    """

    def __init__(
        self, nonconf_score_func: Callable, weight_func: Callable = None
    ):
        self.nonconf_score_func = nonconf_score_func
        self._nonconf_scores = None
        self._calib_len = 0
        self.weight_func = weight_func

    def fit(self, z: Iterable):
        """Compute and store nonconformity scores on the calibration set.

        :param Iterable z: calibration dataset.
        """
        self._nonconf_scores = self.nonconf_score_func(z)
        self._calib_len = len(self._nonconf_scores)

    def set_nonconformity_scores(self, scores: np.array):
        """Setter of nonconformity scores. Can be used instead of calling
        :func:`fit` if the nonconformity scores are already computed.

        :param ndarray scores: nonconformity scores.
        """
        if self._nonconf_scores is not None:
            warnings.warn(
                "Warning........... You are overwriting previously computed or provided scores."
            )
        self._nonconf_scores = np.copy(scores)
        self._calib_len = len(self._nonconf_scores)

    def get_nonconformity_scores(self) -> np.ndarray:
        """Getter of computed nonconformity scores on the calibration set.

        :returns: nonconformity scores.
        :rtype: np.ndarray
        """
        return self._nonconf_scores

    def is_conformal(self, z: Iterable, alpha: float) -> np.ndarray:
        """Test if new data points `z` are conformal. The test result is True
        if the new sample is conformal w.r.t a significance level
        :math:`\\alpha` and False otherwise.

        :param Iterable z: new samples.
        :param float alpha: significance level.

        :returns: conformity test results.
        :rtype: np.ndarray[bool]
        """
        if self._nonconf_scores is None:
            raise RuntimeError(
                "Run `fit` or 'set_nonconformity_scores' methods before calling `is_conformal`."
            )

        n = self._calib_len
        weights = None

        if self.weight_func:
            weights = self.weight_func(z)

        q_hat = quantile(
            self._nonconf_scores, q=(1 - alpha) * (n + 1) / n, w=weights
        )

        test_nonconf_scores = self.nonconf_score_func(z)

        return test_nonconf_scores <= q_hat


class CvPlusCalibrator:
    """Meta calibrator that combines the estimations of nonconformity
    scores by each K-Fold calibrator and produces associated prediction
    intervals based on `CV+ <https://arxiv.org/abs/1905.02928>`_.

    :param dict kfold_calibrators_dict: collection of calibrators for each
        K-fold (disjoint calibration subsets). Each calibrator needs to priorly
        estimate the nonconformity scores w.r.t the associated calibration fold.

    """

    def __init__(self, kfold_calibrators: dict):
        self.kfold_calibrators_dict = kfold_calibrators
        self._len_calib = None
        self._feature_axis = None

        # Sanity checks:
        #   - The collection of calibrators is not None
        if kfold_calibrators is None:
            raise RuntimeError("Calibrators not defined.")

        #   - The calibrators in the collection are not None
        for k, calibrator in self.kfold_calibrators_dict.items():
            if calibrator is None:
                raise RuntimeError(f"Fold {k} calibrator is not defined.")

    def fit(self) -> None:
        """Check if all calibrators have already been fitted.

        :raises RuntimeError: one or more of the calibrators did not estimate
            the nonconformity scores.

        """
        len_calib = 0
        for k, calibrator in self.kfold_calibrators_dict.items():
            kth_nconf_scores = calibrator.get_nonconformity_scores()
            if kth_nconf_scores is None:
                error_msg = (
                    f"Fold {k} calibrator should have priorly "
                    + "been estimated its nonconformity scores."
                )
                raise RuntimeError(error_msg)
            len_calib += len(kth_nconf_scores)
        self._len_calib = len_calib

    def calibrate(
        self,
        *,
        X: Iterable,
        kfold_predictors_dict: dict,
        alpha: float,
    ) -> Tuple[np.ndarray]:
        """Compute calibrated prediction intervals for new examples X.

        :param Iterable X: test features.
        :param dict kfold_predictors_dict: dictionnary of predictors trained
            on each fold.
        :param float alpha: significance level (maximum miscoverage target).

        :returns: y_lower, y_upper.
        :rtype: Tuple[ndarray]
        """
        # Check if all calibrators have already been fitted
        self.fit()

        # Check consistency of alpha w.r.t the size of calibration data
        alpha_calib_check(alpha=alpha, n=self._len_calib)

        # Init the collection of upper and lower bounds of the K-fold's PIs
        concat_y_lo = None
        concat_y_hi = None
        concat_norm_weights = None

        for k, predictor in kfold_predictors_dict.items():
            # Predictions
            y_pred = predictor.predict(X)

            if y_pred is None:
                raise RuntimeError("No prediction obtained with cv+.")

            # Check for multivariate predictions
            if y_pred.ndim > 1:
                self._feature_axis = -1

            # nonconformity scores
            kth_calibrator = self.kfold_calibrators_dict[k]
            nconf_scores = kth_calibrator.get_nonconformity_scores()
            norm_weights = kth_calibrator.get_norm_weights()

            # Reshaping nonconformity scores to broadcast them
            # on y_pred samples when computing the prediction sets
            # Source: R. Barber Section 3 https://arxiv.org/pdf/1905.02928.pdf
            y_pred = y_pred[..., np.newaxis]
            if len(nconf_scores.shape) != 2:
                try:
                    nconf_scores = np.array(nconf_scores)
                    nconf_scores = nconf_scores[np.newaxis, ...]
                except Exception:
                    # @TODO extend the scope beyond castable to ndarrays
                    raise RuntimeError(
                        "Cannot cast nonconformity scores to numpy array."
                    )

            if concat_y_lo is None or concat_y_hi is None:
                (concat_y_lo, concat_y_hi) = kth_calibrator.pred_set_func(
                    y_pred, nconf_scores
                )
                if norm_weights is not None:
                    concat_norm_weights = norm_weights
            else:
                y_lo, y_hi = kth_calibrator.pred_set_func(y_pred, nconf_scores)
                concat_y_lo = np.concatenate([concat_y_lo, y_lo], axis=1)
                concat_y_hi = np.concatenate([concat_y_hi, y_hi], axis=1)
                if norm_weights is not None:
                    concat_norm_weights = np.concatenate(
                        [concat_norm_weights, norm_weights]
                    )

        # sanity check
        if concat_y_lo is None or concat_y_hi is None:
            raise RuntimeError("This should never happen.")

        if concat_norm_weights is None:
            weights = None
        else:
            weights = concat_norm_weights
            infty_array = np.array([np.inf])
            concat_y_lo = np.concatenate([concat_y_lo, infty_array])
            concat_y_hi = np.concatenate([concat_y_hi, infty_array])

        y_lo = -1 * quantile(
            -1 * concat_y_lo,
            (1 - alpha) * (1 + 1 / self._len_calib),
            w=weights,
            axis=1,
            feature_axis=self._feature_axis,
        )
        y_hi = quantile(
            concat_y_hi,
            (1 - alpha) * (1 + 1 / self._len_calib),
            w=weights,
            axis=1,
            feature_axis=self._feature_axis,
        )
        return y_lo, y_hi



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides prediction sets for conformal prediction. To be used when
building a :ref:`calibrator <calibration>`.
"""
import logging
import pkgutil
from typing import Callable
from typing import Iterable
from typing import List
from typing import Tuple

import numpy as np

from deel.puncc.api.utils import logit_normalization_check
from deel.puncc.api.utils import supported_types_check

if pkgutil.find_loader("pandas") is not None:
    import pandas as pd

logger = logging.getLogger(__name__)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Classification ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def lac_set(
        Y_pred, scores_quantile
) -> List:
    """LAC prediction set.

    :param Iterable Y_pred:
        :math:`Y_{\\text{pred}} = (P_{\\text{C}_1}, ..., P_{\\text{C}_n})`
        where :math:`P_{\\text{C}_i}` is logit associated to class i.
        
    :param ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`


    :returns: LAC prediction sets.
    :rtype: Iterable

    """
    # Check if logits sum is close to one
    logit_normalization_check(Y_pred)

    pred_len = len(Y_pred)

    logger.debug(f"Shape of Y_pred: {Y_pred.shape}")

    # Build prediction sets
    prediction_sets = [
        np.where(Y_pred[i] >= 1 - scores_quantile)[0].tolist() for i in range(pred_len)
    ]

    return (prediction_sets,)


def raps_set(
    Y_pred, scores_quantile, lambd: float = 0, k_reg: int = 1, rand: bool = True
) -> List:
    """RAPS prediction set.

    .. warning::

        This signature is incompatible with the interface of calibrators.
        **Use** :func:`raps_set_builder` **to properly initialize**
        :class:`deel.puncc.api.calibration.BaseCalibrator`.

    :param Iterable Y_pred:
        :math:`Y_{\\text{pred}} = (P_{\\text{C}_1}, ..., P_{\\text{C}_n})`
        where :math:`P_{\\text{C}_i}` is logit associated to class i.
        
    :param ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`
        
    :param float lambd: positive weight associated to the regularization term
        that encourages small set sizes. If :math:`\\lambda = 0`, there is no
        regularization and the implementation identifies with **APS**.
        
    :param float k_reg: class rank (ordered by descending probability) starting
        from which the regularization is applied. For example, if
        :math:`k_{reg} = 3`, then the fourth most likely estimated class has an
        extra penalty of size :math:`\\lambda`.
        
    : param bool rand: turn on or off the randomization term that smoothes the
        discrete probability mass jump when including a new class.


    :returns: RAPS prediction sets.
    :rtype: Iterable

    """
    # Check if logits sum is close to one
    logit_normalization_check(Y_pred)

    pred_len = len(Y_pred)

    logger.debug(f"Shape of Y_pred: {Y_pred.shape}")

    if rand:
        # Generate u randomly from a uniform distribution
        u = np.random.uniform(size=pred_len)

    # 1-alpha th empirical quantile of conformity scores
    tau = scores_quantile

    # Rank classes by their probability
    idx_class_pred_ranking = np.argsort(-Y_pred, axis=-1)
    sorted_proba = -np.sort(-Y_pred, axis=-1)
    sorted_cum_mass = sorted_proba.cumsum(axis=-1)

    # Cumulative probability mass of (sorted by descending probability) classes
    # penalized by the regularization term
    penal_cum_proba = sorted_cum_mass + lambd * np.maximum(
        np.arange(1, sorted_cum_mass.shape[-1] + 1) - k_reg, 0
    )

    # L is the number of classes (+1) for which the cumulative probability mass
    # is below the threshold "tau"
    # The minimum is used in case the Y_pred logits are not well normalized
    L = np.minimum(np.sum(penal_cum_proba < tau, axis=-1) + 1, pred_len)

    if not rand:
        # Build prediction set
        prediction_sets = [
            list(idx_class_pred_ranking[i, : L[i]]) for i in range(pred_len)
        ]
        return (prediction_sets,)

    ## The following code runs only when rand argument is True
    # For indexing, use L-1 to denote the L-th element
    # Residual of cumulative probability mass (regularized) and tau
    proba_excess = (
        sorted_cum_mass[np.arange(pred_len), L - 1]
        + lambd * np.maximum(L - k_reg, 0)
        - tau
    )

    # Indicator when regularization rank is exceeded
    indic_L_greater_kreg = np.where(L > k_reg, 1, 0)

    # Normalized probability mass excess
    v = proba_excess / (
        sorted_proba[np.arange(pred_len), L - 1] + lambd * indic_L_greater_kreg
    )

    # Least likely class in the prediction set is removed
    # with a probability of norm_proba_excess
    indic_excess_proba = np.where(u <= v, 1, 0)
    L = L - indic_excess_proba

    # Build prediction set
    prediction_sets = [
        list(idx_class_pred_ranking[i, : L[i]]) for i in range(pred_len)
    ]

    return (prediction_sets,)


def raps_set_builder(
    lambd: float = 0, k_reg: int = 1, rand: bool = True
) -> Callable:
    """RAPS prediction set builder. When called, returns a RAPS prediction set
    function :func:`raps_set` with given initialitation of regularization
    hyperparameters.

    :param float lambd: positive weight associated to the regularization term
        that encourages small set sizes. If :math:`\\lambda = 0`, there is no
        regularization and the implementation identifies with **APS**.
        
    :param float k_reg: class rank (ordered by descending probability) starting
        from which the regularization is applied. For example, if
        :math:`k_{reg} = 3`, then the fourth most likely estimated class has an
        extra penalty of size :math:`\\lambda`.
        
    : param bool rand: turn on or off the randomization term that smoothes the
        discrete probability mass jump when including a new class.

    :returns: RAPS prediction set function that takes two parameters:
        `Y_pred` and `scores_quantile`.
    :rtype: Callable

    :raises ValueError: incorrect value of lambd or k_reg.
    :raises TypeError: unsupported data types.
    
    """
    if lambd < 0:
        raise ValueError(
            f"Argument `lambd` has to be positive, provided: {lambd} < 0"
        )
    if k_reg < 0:
        raise ValueError(
            f"Argument `k_reg` has to be positive, provided: {k_reg} < 0"
        )
    # @TODO: type checking and co

    def _raps_set_function(Y_pred, scores_quantile):
        return raps_set(
            Y_pred, scores_quantile, lambd=lambd, k_reg=k_reg, rand=rand
        )

    return _raps_set_function


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Regression ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def constant_interval(
    y_pred: Iterable, scores_quantile: np.ndarray
) -> Tuple[np.ndarray]:
    """Constant prediction interval centered on `y_pred`. The size of the
    margin is `scores_quantile` (noted :math:`\gamma_{\\alpha}`).

    .. math::

        I = [y_{\\text{pred}} - \gamma_{\\alpha}, y_{\\text{pred}} +
        \gamma_{\\alpha}]

    :param Iterable y_pred: predictions.
    
    :param ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`.

    :returns: prediction intervals :math:`I`.
    :rtype: Tuple[ndarray]

    :raises TypeError: unsupported data types.
    """
    supported_types_check(y_pred)
    y_lo = y_pred - scores_quantile
    y_hi = y_pred + scores_quantile
    return y_lo, y_hi


def scaled_interval(
    Y_pred: Iterable, scores_quantile: np.ndarray, eps: float = 1e-12
) -> Tuple[np.ndarray]:
    """Scaled prediction interval centered on `y_pred`. Considering
    :math:`Y_{\\text{pred}} = (\mu_{\\text{pred}}, \sigma_{\\text{pred}})`,
    the size of the margin is proportional to `scores_quantile`
    :math:`\gamma_{\\alpha}`.

    .. math::

        I = [\mu_{\\text{pred}} - \gamma_{\\alpha} \cdot \sigma_{\\text{pred}},
        y_{\\text{pred}} + \gamma_{\\alpha} \cdot \sigma_{\\text{pred}}]

    :param Iterable y_pred: predictions.
    :param ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`.
    :param float eps: small positive value to avoid singleton sets.

    :returns: scaled prediction intervals :math:`I`.
    :rtype: Tuple[ndarray]

    :raises TypeError: unsupported data types.
    """
    supported_types_check(Y_pred)

    if Y_pred.shape[1] != 2:  # check Y_pred contains two predictions
        raise RuntimeError(
            "Each Y_pred must contain a point prediction and a dispersion estimation."
        )

    if pkgutil.find_loader("pandas") is not None and isinstance(
        Y_pred, pd.DataFrame
    ):
        y_pred, sigma_pred = Y_pred.iloc[:, 0], Y_pred.iloc[:, 1]
    else:
        y_pred, sigma_pred = Y_pred[:, 0], Y_pred[:, 1]

    if np.any(sigma_pred + eps <= 0):
        print("Warning: test points with MAD predictions below -eps"
              " will have infinite sized prediction intervals.")

    fints = sigma_pred + eps > 0
    y_lo, y_hi = np.zeros_like(y_pred), np.zeros_like(y_pred)
    y_lo[fints] = y_pred[fints] - scores_quantile * (sigma_pred[fints] + eps)
    y_hi[fints] = y_pred[fints] + scores_quantile * (sigma_pred[fints] + eps)
    y_lo[~fints] = -np.inf
    y_hi[~fints] = np.inf
    return y_lo, y_hi


def cqr_interval(
    Y_pred: Iterable, scores_quantile: np.ndarray
) -> Tuple[np.ndarray]:
    """CQR prediction interval. Considering
    :math:`Y_{\\text{pred}} = (q_{\\text{lo}}, q_{\\text{hi}})`, the prediction
    interval is built from the upper and lower quantiles predictions and
    `scores_quantile` :math:`\gamma_{\\alpha}`.

    .. math::

        I = [q_{\\text{lo}} - \gamma_{\\alpha}, q_{\\text{lo}} +
        \gamma_{\\alpha}]

    :param Iterable y_pred: predictions.
    :param ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`.

    :returns: scaled prediction intervals :math:`I`.
    :rtype: Tuple[ndarray]

    :raises TypeError: unsupported data types.
    """
    supported_types_check(Y_pred)

    if Y_pred.shape[1] != 2:  # check Y_pred contains two predictions
        raise RuntimeError(
            "Each Y_pred must contain lower and higher quantiles predictions, respectively."
        )

    if pkgutil.find_loader("pandas") is not None and isinstance(
        Y_pred, pd.DataFrame
    ):
        q_lo, q_hi = Y_pred.iloc[:, 0], Y_pred.iloc[:, 1]
    else:
        q_lo, q_hi = Y_pred[:, 0], Y_pred[:, 1]

    y_lo = q_lo - scores_quantile
    y_hi = q_hi + scores_quantile
    return y_lo, y_hi


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Object Detection ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def constant_bbox(Y_pred: np.ndarray, scores_quantile: np.ndarray):
    """
    Generate the upper and lower bounds of the bounding box coordinates for
    a given prediction.

    :param np.ndarray Y_pred: the predicted bounding box coordinates.
    :param np.ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`.

    :return: the lower bound and upper bound coordinates of the bounding box.
    :rtype: Tuple[np.ndarray, np.ndarray]

    :raises TypeError: unsupported data types.
    """

    if not isinstance(Y_pred, np.ndarray):
        raise TypeError(
            f"Unsupported data type {type(Y_pred)}."
            + "Please provide a numpy ndarray"
        )

    if Y_pred.shape[1] != 4:  # check Y_pred contains four coordinates
        raise RuntimeError("Each Y_pred must contain 4 bbox coordinates.")

    # Recover each coordinate
    x_min, y_min, x_max, y_max = np.hsplit(Y_pred, 4)

    # Coordinates of covering bbox (upperbounds)
    x_min_lo, y_min_lo = x_min - scores_quantile[0], y_min - scores_quantile[1]
    x_max_hi, y_max_hi = x_max + scores_quantile[2], y_max + scores_quantile[3]
    Y_pred_hi = np.hstack([x_min_lo, y_min_lo, x_max_hi, y_max_hi])

    # Coordinates of included bbox (lowerbounds)
    x_min_hi, y_min_hi = x_min + scores_quantile[0], y_min + scores_quantile[1]
    x_max_lo, y_max_lo = x_max - scores_quantile[2], y_max - scores_quantile[3]
    Y_pred_lo = np.hstack([x_min_hi, y_min_hi, x_max_lo, y_max_lo])

    return Y_pred_lo, Y_pred_hi


def scaled_bbox(Y_pred: np.ndarray, scores_quantile: np.ndarray):
    """
    Scaled upper and lower bounds of the bounding box coordinates
    for a given prediction coordinates.

    :param np.ndarray Y_pred: the predicted bounding box coordinates.
    :param np.ndarray scores_quantile: quantile of nonconformity scores computed
        on a calibration set for a given :math:`\\alpha`.

    :return: The coordinates of the inner and outer bounding boxes.
    :rtype: Tuple[np.ndarray, np.ndarray]

    :raises TypeError: unsupported data types.
    """
    if not isinstance(Y_pred, np.ndarray):
        raise TypeError(
            f"Unsupported data type {type(Y_pred)}."
            + "Please provide a numpy ndarray"
        )

    if Y_pred.shape[1] != 4:  # check Y_pred contains four coordinates
        raise RuntimeError("Each Y_pred must contain 4 bbox coordinates.")

    # Recover each coordinate
    x_min, y_min, x_max, y_max = np.hsplit(Y_pred, 4)

    # Compute width and height of predicted bbox
    dx = np.abs(x_max - x_min)
    dy = np.abs(y_max - y_min)

    # Coordinates of covering bbox (upperbounds)
    x_min_lo = x_min - scores_quantile[0] * dx
    y_min_lo = y_min - scores_quantile[1] * dy
    x_max_hi = x_max + scores_quantile[2] * dx
    y_max_hi = y_max + scores_quantile[3] * dy

    Y_pred_outer = np.hstack([x_min_lo, y_min_lo, x_max_hi, y_max_hi])

    # Coordinates of included bbox (lowerbounds)
    x_min_hi, y_min_hi = (
        x_min + scores_quantile[0] * dx,
        y_min + scores_quantile[1] * dy,
    )
    x_max_lo, y_max_lo = (
        x_max - scores_quantile[2] * dx,
        y_max - scores_quantile[3] * dy,
    )
    Y_pred_inner = np.hstack([x_min_hi, y_min_hi, x_max_lo, y_max_lo])

    return Y_pred_inner, Y_pred_outer



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides nonconformity scores for conformal prediction. To be used
when building a :ref:`calibrator <calibration>`.
"""
import importlib
from typing import Callable
from typing import Iterable

import numpy as np

from deel.puncc.api.utils import logit_normalization_check
from deel.puncc.api.utils import supported_types_check

if importlib.util.find_spec("pandas") is not None:
    import pandas as pd

if importlib.util.find_spec("tensorflow") is not None:
    import tensorflow as tf

if importlib.util.find_spec("torch") is not None:
    import torch


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Classification ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def lac_score(
        Y_pred: Iterable,
        y_true: Iterable,
) -> Iterable:
    """LAC nonconformity score.

    :param Iterable Y_pred:
        :math:`Y_{\\text{pred}} = (P_{\\text{C}_1}, ..., P_{\\text{C}_n})`
        where :math:`P_{\\text{C}_i}` is logit associated to class i.
    :param Iterable y_true: true labels.

    :returns: RAPS nonconformity scores.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """
    supported_types_check(Y_pred, y_true)

    # Check if logits sum is close to one
    logit_normalization_check(Y_pred)

    if not isinstance(Y_pred, np.ndarray):
        raise NotImplementedError(
            "LAC nonconformity score only implemented for ndarrays"
        )

    # Compute and return the LAC nonconformity score
    return 1 - Y_pred[np.arange(y_true.shape[0]), y_true]


def raps_score(
    Y_pred: Iterable,
    y_true: Iterable,
    lambd: float = 0,
    k_reg: int = 1,
    rand: bool = True,
) -> Iterable:
    """RAPS nonconformity score.

    .. warning::

        This signature is incompatible with the interface of calibrators.
        **Use** :func:`raps_score_builder` **to properly initialize**
        :class:`deel.puncc.api.calibration.BaseCalibrator`.

    :param Iterable Y_pred:
        :math:`Y_{\\text{pred}} = (P_{\\text{C}_1}, ..., P_{\\text{C}_n})`
        where :math:`P_{\\text{C}_i}` is logit associated to class i.
    :param Iterable y_true: true labels.
    :param float lambd: positive weight associated to the regularization term
        that encourages small set sizes. If :math:`\\lambda = 0`, there is no
        regularization and the implementation identifies with **APS**.
    :param float k_reg: class rank (ordered by descending probability) starting
        from which the regularization is applied. For example,
        if :math:`k_{reg} = 3`, then the fourth most likely estimated class has
        an extra penalty of size :math:`\\lambda`.
    :param bool rand: turn on or off the randomization term that smoothes the
        discrete probability mass jump when including a new class.


    :returns: RAPS nonconformity scores.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """
    supported_types_check(Y_pred, y_true)

    # Check if logits sum is close to one
    logit_normalization_check(Y_pred)

    if not isinstance(Y_pred, np.ndarray):
        raise NotImplementedError(
            "RAPS/APS nonconformity scores only implemented for ndarrays"
        )

    # Sort classes by descending probability order
    class_ranking = np.argsort(-Y_pred, axis=1)
    sorted_proba = -np.sort(-Y_pred, axis=1)

    # Cumulative probability mass (given the previous class ranking)
    sorted_cum_mass = sorted_proba.cumsum(axis=1)

    # Locate position of true label in the classes
    # sequence ranked by decreasing probability
    L = [
        np.where(class_ranking[i] == y_true[i])[0][0]
        for i in range(y_true.shape[0])
    ]

    # Threshold of cumulative probability mass to include the real class
    E = [sorted_cum_mass[i, L[i]] for i in range(y_true.shape[0])]

    if rand:
        # Generate u randomly from a uniform distribution
        u = np.random.uniform(size=len(y_true))
        E = [
            E[i]
            + lambd * np.maximum((L[i] + 1 - k_reg), 0)
            - u[i] * sorted_proba[i, L[i]]
            for i in range(y_true.shape[0])
        ]
    else:
        E = [
            E[i] + lambd * np.maximum((L[i] + 1 - k_reg), 0)
            for i in range(y_true.shape[0])
        ]

    return np.array(E)


def raps_score_builder(
    lambd: float = 0, k_reg: int = 1, rand: bool = True
) -> Callable:
    """RAPS nonconformity score builder. When called, returns a RAPS
    nonconformity score function :func:`raps_score` with given initialitation
    of regularization hyperparameters.

    :param float lambd: positive weight associated to the regularization term
        that encourages small set sizes. If :math:`\\lambda = 0`, there is no
        regularization and the implementation identifies with **APS**.
    :param float k_reg: class rank (ordered by descending probability) starting
        from which the regularization is applied. For example, if
        :math:`k_{reg} = 3`, then the fourth most likely estimated class has
        an extra penalty of size :math:`\\lambda`.
    :param bool rand: turn on or off the randomization term that smoothes the
        discrete probability mass jump when including a new class.

    :returns: RAPS nonconformity score function that takes two parameters:
        `Y_pred` and `y_true`.
    :rtype: Callable

    :raises TypeError: unsupported data types.
    """

    def _raps_score_function(Y_pred: Iterable, y_true: Iterable) -> np.ndarray:
        return raps_score(Y_pred, y_true, lambd, k_reg, rand)

    return _raps_score_function


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Regression ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def difference(y_pred: Iterable, y_true: Iterable) -> Iterable:
    """Elementwise difference.

    .. math::

        R = y_{\\text{pred}}-y_{\\text{true}}

    :param Iterable y_pred: predictions.
    :param Iterable y_true: true labels.

    :returns: coordinatewise difference.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """
    supported_types_check(y_pred, y_true)

    if importlib.util.find_spec("torch") is not None and isinstance(
        y_pred, torch.Tensor
    ):
        y_pred = y_pred.cpu().detach().numpy()
        y_true = y_true.cpu().detach().numpy()
        return np.squeeze(y_pred) - np.squeeze(y_true)

    return y_pred - y_true


def absolute_difference(y_pred: Iterable, y_true: Iterable) -> Iterable:
    """Absolute Deviation.

    .. math::

        R = |y_{\\text{true}}-y_{\\text{pred}}|

    :param Iterable y_pred: predictions.
    :param Iterable y_true: true labels.

    :returns: mean absolute deviation.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """

    return abs(difference(y_pred, y_true))


def scaled_ad(
    Y_pred: Iterable, y_true: Iterable, eps: float = 1e-12
) -> Iterable:
    """Scaled Absolute Deviation, normalized by an estimation of the conditional
    mean absolute deviation (conditional MAD). Considering
    :math:`Y_{\\text{pred}} = (\mu_{\\text{pred}}, \sigma_{\\text{pred}})`:

    .. math::

        R = \\frac{|y_{\\text{true}}-\mu_{\\text{pred}}|}{\sigma_{\\text{pred}}}

    :param Iterable Y_pred: point and conditional MAD predictions.
        :math:`Y_{\\text{pred}}=(y_{\\text{pred}}, \sigma_{\\text{pred}})`
    :param Iterable y_true: true labels.
    :param float eps: small positive value to avoid division by negative or zero.

    :returns: scaled absolute deviation.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """
    supported_types_check(Y_pred, y_true)

    if Y_pred.shape[1] != 2:  # check Y_pred contains two predictions
        raise RuntimeError(
            "Each Y_pred must contain a point prediction and a dispersion estimation."
        )

    # check y_true is a collection of point observations
    if len(y_true.shape) != 1:
        raise RuntimeError("Each y_true must contain a point observation.")

    if importlib.util.find_spec("pandas") is not None and isinstance(
        Y_pred, pd.DataFrame
    ):
        y_pred, sigma_pred = Y_pred.iloc[:, 0], Y_pred.iloc[:, 1]
    else:
        y_pred, sigma_pred = Y_pred[:, 0], Y_pred[:, 1]

    # MAD then Scaled MAD and computed
    mean_absolute_deviation = absolute_difference(y_pred, y_true)
    if np.any(sigma_pred + eps <= 0):
        print(
            "Warning: calibration points with MAD predictions"
            " below -eps won't be used for calibration."
        )

    nonneg = sigma_pred + eps > 0
    return mean_absolute_deviation[nonneg] / (sigma_pred[nonneg] + eps)


def cqr_score(Y_pred: Iterable, y_true: Iterable) -> Iterable:
    """CQR nonconformity score. Considering
    :math:`Y_{\\text{pred}} = (q_{\\text{lo}}, q_{\\text{hi}})`:

    .. math::

        R = max\{q_{\\text{lo}} - y_{\\text{true}}, y_{\\text{true}} - q_{\\text{hi}}\}

    where :math:`q_{\\text{lo}}` (resp. :math:`q_{\\text{hi}}`) is the lower
    (resp. higher) quantile prediction

    :param Iterable Y_pred: predicted quantiles couples.
    :param Iterable y_true: true quantiles couples.

    :returns: CQR nonconformity scores.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """
    supported_types_check(Y_pred, y_true)

    if Y_pred.shape[1] != 2:  # check Y_pred contains two predictions
        raise RuntimeError(
            "Each Y_pred must contain lower and higher quantiles estimations."
        )

    # check y_true is a collection of point observations
    if len(y_true.shape) != 1:
        raise RuntimeError("Each y_pred must contain a point observation.")

    if importlib.util.find_spec("pandas") is not None and isinstance(
        Y_pred, pd.DataFrame
    ):
        q_lo, q_hi = Y_pred.iloc[:, 0], Y_pred.iloc[:, 1]
    else:
        q_lo, q_hi = Y_pred[:, 0], Y_pred[:, 1]

    diff_lo = q_lo - y_true
    diff_hi = y_true - q_hi

    if isinstance(diff_lo, np.ndarray):
        return np.maximum(diff_lo, diff_hi)

    if importlib.util.find_spec("pandas") is not None and isinstance(
        diff_lo, (pd.DataFrame, pd.Series)
    ):
        return (pd.concat([diff_lo, diff_hi]).groupby(level=0)).max()
        # raise NotImplementedError(
        #     "CQR score not implemented for DataFrames. Please provide ndarray or tensors."
        # )

    if importlib.util.find_spec("tensorflow") is not None and isinstance(
        diff_lo, tf.Tensor
    ):
        return tf.math.maximum(diff_lo, diff_hi)

    # if importlib.util.find_spec("torch") is not None and isinstance(
    #     diff_lo, torch.Tensor
    # ):
    #     return torch.maximum(diff_lo, diff_hi)

    raise RuntimeError("Fatal Error. Type check failed !")


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Object Detection ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def scaled_bbox_difference(Y_pred: np.ndarray, Y_true: np.ndarray):
    """Object detection scaled nonconformity score. Considering
    :math:`Y_{\\text{pred}} = (\hat{x}_1, \hat{y}_1, \hat{x}_2, \hat{y}_2)` and
    :math:`Y_{\\text{true}} = (x_1, y_1, x_2, y_2)`:

    .. math::

        R = (\\frac{\hat{x}_1-x_1}{\Delta x}, \\frac{\hat{y}_1-y_1}{\Delta y},
        \\frac{\hat{x}_2-x_2}{\Delta x}, \\frac{\hat{y}_1-y_1}{\Delta y})

    where :math:`\Delta x = |\hat{x}_2 - \hat{x}_1|` and
    :math:`\Delta y = |\hat{y}_2 - \hat{y}_1|`.

    :param Iterable Y_pred: predicted coordinates of the bounding box.
    :param Iterable y_true: true coordinates of the bounding box.

    :returns: scaled object detection nonconformity scores.
    :rtype: Iterable

    :raises TypeError: unsupported data types.
    """
    if not isinstance(Y_pred, np.ndarray):
        raise TypeError(
            f"Unsupported data type {type(Y_pred)}."
            + "Please provide a numpy ndarray"
        )

    if Y_pred.shape[1] != 4:  # check Y_pred contains four coordinates
        raise RuntimeError("Each Y_pred must contain 4 bbox coordinates.")

    # check y_true is formatted as a bbox with 4 coordinates
    if Y_true.shape[1] != 4:
        raise RuntimeError("Each Y_true must contain 4 bbox coordinates.")

    x_min, y_min, x_max, y_max = np.hsplit(Y_pred, 4)
    dx = np.abs(x_max - x_min)
    dy = np.abs(y_max - y_min)
    return (Y_pred - Y_true) / np.hstack([dx, dy, dx, dy])



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements utility functions.
"""
import logging
import importlib
import sys
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union
from scipy.optimize import linear_sum_assignment

from deel.puncc.metrics import iou

import numpy as np

if importlib.util.find_spec("pandas") is not None:
    import pandas as pd

if importlib.util.find_spec("tensorflow") is not None:
    import tensorflow as tf

if importlib.util.find_spec("torch") is not None:
    import torch

logger = logging.getLogger(__name__)

EPSILON = sys.float_info.min  # small value to avoid underflow


def dual_predictor_check(l: Any, name: str, ltype: str):
    """Check if properties of `DualPredictor` come in
    couples.

    :param Any l: property of `DualPredictor`.
    :param str name: name of the property.
    :param str ltype: type of property.

    :raises TypeError: when the properties count is not two.

    """
    if (not isinstance(l, list)) or len(l) != 2:
        raise TypeError(
            f"Argument `{name}` should be a list of two {ltype}. Provided {l}."
        )


def logit_normalization_check(y: Iterable):
    """Check if provided logits sum is close to one.

    :param Iterable y: logits array. Rows correspond to samples and columns to
        classes.

    :raises ValueError: when logits sum is different than one within a tolerance
        value of 1e-5.
    """
    logits_sum = np.sum(np.array(list(y)), -1)
    if np.any(np.abs(logits_sum - 1) > 1e-5):
        raise ValueError(
            f"Logits must some to 1. Provided logit array {logits_sum}"
        )


def sample_len_check(a: Iterable, b: Iterable):
    """Check if arguments type have the same length.

    :param Iterable a: iterable whose type is supported.
    :param Iterable b: iterable whose type is supported.

    :raises TypeError: unsupported data types or elements have inconsistent types.
    :raises ValueError: arguments contain different number of samples.
    """
    supported_types_check(a, b)
    if a.shape[0] != b.shape[0]:
        raise ValueError("Iterables must contain the same number of samples.")


def features_len_check(a: Iterable, b: Iterable):
    """Check if arguments have the same number of features,
    that is their last axes have the same length.

    :param Iterable a: iterable whose type is supported.
    :param Iterable b: iterable whose type is supported.

    :raises TypeError: unsupported data types or elements have inconsistent types.
    :raises ValueError: arguments have different number of features
    """

    supported_types_check(a, b)

    if a.shape[-1] != b.shape[-1]:
        raise ValueError(
            "X_fit and X_calib must contain the same number of features."
        )


def supported_types_check(*data: Iterable):
    """Check if arguments' types are supported.

    :param Iterable data: iterable(s) to be checked.

    :raises TypeError: unsupported data types.
    """

    for a in data:
        if isinstance(a, np.ndarray):
            pass

        elif importlib.util.find_spec("pandas") is not None and isinstance(
            a, (pd.DataFrame, pd.Series)
        ):
            pass
        elif importlib.util.find_spec("tensorflow") is not None and isinstance(
            a, tf.Tensor
        ):
            pass
        elif importlib.util.find_spec("torch") is not None and isinstance(
            a, torch.Tensor
        ):
            pass
        else:
            raise TypeError(
                f"Unsupported data type {type(a)}. Please provide a numpy ndarray, "
                "a dataframe or a tensor."
            )


def alpha_calib_check(
    alpha: Union[float, np.ndarray], n: int, complement_check: bool = False
):
    """Check if the value of alpha :math:`\\alpha` is consistent with the size
    of calibration set :math:`n`.

    The quantile order is inflated by a factor :math:`(1+1/n)` and has to be in
    the interval (0,1). From this, we derive the condition:

    .. math::

        0 < (1-\\alpha)\cdot(1+1/n) < 1 \implies 1 > \\alpha > 1/(n+1)

    If complement_check is set, we consider an additional condition:

    .. math::

        0 < \\alpha \cdot (1+1/n) < 1 \implies 0 < \\alpha < n/(n+1)

    :param float or np.ndarray alpha: target quantile order.
    :param int n: size of the calibration dataset.
    :param bool complement_check: complementary check to compute the
        :math:`\\alpha \cdot (1+1/n)`-th quantile, required by some methods
        such as jk+.

    :raises ValueError: when :math:`\\alpha` is inconsistent with the size of
        calibration set.
    """

    if np.any(alpha < 1 / (n + 1)):
        raise ValueError(
            f"Alpha is too small: (alpha={alpha}, n={n}) "
            + f"{alpha} < {1/(n+1)} for at least one of its coordinates. "
            + "Increase alpha or the size of the calibration set."
        )

    if np.any(alpha >= 1):
        raise ValueError(
            f"Alpha={alpha} is too large. "
            + "Decrease alpha such that all of its coordinates are < 1."
        )

    if complement_check and np.any(alpha > n / (n + 1)):
        raise ValueError(
            f"Alpha is too large: (alpha={alpha}, n={n}) "
            + f"{alpha} < {n/(n+1)} for at least on of its coordinates. "
            + "Decrease alpha or increase the size of the calibration set."
        )

    if complement_check and np.any(alpha <= 0):
        raise ValueError(
            f"Alpha={alpha} is too small. "
            + "Increase alpha such that all of its coordinates > 0."
        )


def get_min_max_alpha_calib(
    n: int, two_sided_conformalization: bool = False
) -> Optional[Tuple[float]]:
    """Get the greatest alpha in (0,1) that is consistent with the size of
    calibration set. Recall that while true Conformal Prediction
    (CP, Vovk et al 2005) is defined only on (0,1) (bounds not included), there
    maybe cases where we must handle alpha=0 or 1 (e.g. Adaptive Conformal
    Inference, Gibbs & Candès 2021).

    This function computes the admissible range of alpha for split CP
    (Papadopoulos et al 2002, Lei et al 2018) and the Jackknife-plus
    (Barber et al 2019). For more CP method, see the literature and update this
    function.

    In Split Conformal prediction we take the index
    := ceiling( (1-alpha)(n+1) )-th element from the __sorted__ nonconformity
    scores as the margin of error. We must have
    1 <= index <= n ( = len(calibration_set)) which boils down to:
    1/(n+1) <= alpha < 1.

    In the case of cv-plus and jackknife-plus (Barber et al 2019), we need:
        1. ceiling( (1-alpha)(n+1) )-th element of scores
        2. floor(     (alpha)(n+1) )-th element of scores

    The argument two_sided_conformalization=True, ensures the indexes are within
    range for this special case: 1/(n+1) <= alpha <= n/(n+1)

    :param int n: size of the calibration dataset.
    :param bool two_sided_conformalization: alpha threshold for two-sided
        quantile of jackknife+ (Barber et al 2021).

    :returns: lower and upper bounds for alpha (miscoverage probability)
    :rtype: Tuple[float]

    :raises ValueError: must have integer n, boolean two_sided_conformalization
        and n>=1
    """

    if n < 1:
        raise ValueError(f"Invalid input: you need n>=1 but received n={n}")

    if two_sided_conformalization is True:
        return (1 / (n + 1), n / (n + 1))

    # Base case: split conformal prediction (e.g. Lei et al 2019)
    return (1 / (n + 1), 1)


def quantile(
    a: Iterable,
    q: Union[float, np.ndarray],
    w: np.ndarray = None,
    axis: int = None,
    feature_axis: int = None,
) -> Union[float, np.ndarray]:
    """Estimate the columnwise q-th empirical weighted quantiles.

    :param Iterable a: collection of n samples
    :param Union[float, np.ndarray] q: q-th quantiles to compute. All elements must be in (0, 1).
    :param ndarray w: vector of size n. By default, w is None and equal weights
        (:math:`1/n`) are associated.
    :param int axis: axis along which to compute quantiles. If None,
        quantiles are computed along the flattened array.
    :param int feature_axis: if multidim quantile, feature_axis is the axis corresponding
        to the features.

    :raises ValueError: all coordinates of q must be in (0, 1).

    :returns: weighted empirical quantiles.
    :rtype: Union[float, np.ndarray]
    """
    # type checks:
    supported_types_check(a)

    if isinstance(a, np.ndarray):
        pass
    elif importlib.util.find_spec("pandas") is not None and isinstance(
        a, pd.DataFrame
    ):
        a = a.to_numpy()
    elif importlib.util.find_spec("tensorflow") is not None and isinstance(
        a, tf.Tensor
    ):
        a = a.numpy()
    # elif importlib.util.find_spec("torch") is not None:
    #     if isinstance(a, torch.Tensor):
    #         a = a.cpu().detach().numpy()
    else:
        raise RuntimeError(
            "Unsupported data type for argument [a].\n"
            "Please provide a numpy ndarray, a pandas dataframe or a tensorflow tensor."
        )

    # Sanity checks
    if np.any(q <= 0) or np.any(q >= 1):
        raise ValueError(
            "All coordinates of [q] must be in the open interval (0, 1)."
        )

    # Unweighted case
    if w is None:
        return quantile_unweighted(a, q, axis=axis, feature_axis=feature_axis)

    # Weighted case
    return quantile_weighted(a, q, w, axis=axis, feature_axis=feature_axis)


def quantile_unweighted(
    a: np.ndarray,
    q: Union[float, np.ndarray],
    axis: int = None,
    feature_axis: int = None,
) -> Union[float, np.ndarray]:
    """Estimate the multi-dimensional q-th empirical quantiles.

    :param ndarray a: collection of n samples
    :param Union[float, np.ndarray] q: q-th quantiles to compute. All elements must be in (0, 1).
    :param int axis: axis along which to compute quantiles. If None,
        quantiles are computed along all axes except the feature_axis.
    :param int feature_axis: if multidim quantile, feature_axis is the axis
        wich corresponds to the features.

    :raises ValueError: axis value cannot coincide with features axis.
    :raises ValueError: a and q must have the same number of features if q is an array.

    :returns: empirical quantiles.
    :rtype: Union[float, np.ndarray].
    """
    if axis is not None and axis == feature_axis:
        raise ValueError("axis value cannot coincide with features axis.")

    if isinstance(q, float):
        return np.quantile(a, q, axis=axis, method="inverted_cdf")

    if a.shape[feature_axis] != len(q):
        raise ValueError("a and q must have the same number of features.")
    quantile_res = np.array(
        [
            np.quantile(
                np.expand_dims(a.take(i, axis=feature_axis), axis=feature_axis),
                q[i],
                axis=axis,
                method="inverted_cdf",
            )
            for i in range(len(q))
        ]
    )
    return np.squeeze(
        np.transpose(quantile_res, (*range(1, quantile_res.ndim), 0))
    )


def quantile_weighted_unidim(
    a: np.ndarray, q: float, w: np.ndarray, axis=None
) -> Union[float, np.ndarray]:
    """Estimate the one-dimensional weighted q-th empirical quantiles.

    :param ndarray a: collection of n samples
    :param float: q-th quantiles to compute. All elements must be in (0, 1).
    :param ndarray w: array of weights.
    :param int axis: axis along which to compute quantiles. If None,
        quantiles are computed along the flattened array.

    :raises ValueError: w must be a 1D array.
    :raises ValueError: a and w must have the same length.

    :returns: empirical weighted quantiles.
    :rtype: Union[float, np.ndarray].
    """
    # Dimension checks
    if w.ndim != 1:
        raise ValueError("w must be a 1D array.")

    if len(w) != len(a):
        raise ValueError("a and w must have the same length.")

    # Normalization check
    norm_condition = np.isclose(np.sum(w) - 1, 0, atol=1e-14)
    if ~norm_condition:
        error = f"w is not normalized. Sum of weights is {np.sum(w)}"
        raise RuntimeError(error)

    # Values are sorted in ascending order
    sorted_idx = np.argsort(a, axis=axis)
    logger.debug("Sorted indices: %s", sorted_idx)

    # Reorder weights (ascending values of axis of a) and compute cumulative sum
    sorted_cumsum_weights = np.cumsum(w[sorted_idx], axis=axis)
    logger.debug("Sorted weights cumulative sum: %s", sorted_cumsum_weights)

    # Get the smallest index for which the cumulative sum of weights exceeds p
    min_idx_reaching_q = np.sum(
        sorted_cumsum_weights < q, axis=axis, keepdims=True
    )
    logger.debug(
        "First index per column where cumsum exceeds q: %s", min_idx_reaching_q
    )

    # Sort a
    sorted_a = np.take_along_axis(a, sorted_idx, axis=axis)

    # Collect the p-th quantile (first value whose probability mass exceeds p)
    quantile_res = np.take_along_axis(sorted_a, min_idx_reaching_q, axis=axis)
    logger.debug("Quantiles array: %s", quantile_res)
    return np.squeeze(quantile_res)


def quantile_weighted(
    a: np.ndarray,
    q: Union[float, np.ndarray],
    w: np.ndarray,
    axis: int = None,
    feature_axis: int = None,
) -> Union[float, np.ndarray]:
    """Estimate the multi-dimensional weighted q-th empirical quantiles.

    :param ndarray a: collection of n samples
    :param Union[float, np.ndarray] q: q-th quantiles to compute. All elements must be in (0, 1).
    :param ndarray w: array of weights.
    :param int axis: axis along which to compute quantiles. If None,
        quantiles are computed along the flattened array.
    :param int feature_axis: if multidim quantile, feature_axis is the axis
        wich corresponds to the features.

    :raises ValueError: cannot take quantiles along features axis.
    :raises ValueError: w must be have the same number of elements as a along axis.
    :raises ValueError: a and q must have the same number of features if q is an array.
    :raises ValueError: w and q must have the same number of features if w is a 2D array.

    :returns: empirical weighted quantiles.
    :rtype: Union[float, np.ndarray].
    """
    if axis is not None and axis == feature_axis:
        raise ValueError("axis value cannot coincide with features axis.")

    if axis is not None and len(w) != a.shape[axis]:
        raise ValueError("w must have the same length as a.shape[axis].")

    if isinstance(q, float) and w.ndim == 1:
        return quantile_weighted_unidim(a, q, w, axis=axis)

    if isinstance(q, float):
        quantile_res = np.array(
            [
                quantile_weighted_unidim(a[..., i], q, w[:, i], axis=axis)
                for i in range(a.shape[-1])
            ]
        )
    elif a.shape[-1] != len(q):
        raise ValueError("a and q must have the same number of features.")

    if w.ndim == 1:
        quantile_res = np.array(
            [
                quantile_weighted_unidim(a[..., i], q[i], w, axis=axis)
                for i in range(len(q))
            ]
        )
    elif w.ndim == 2 and w.shape[1] != len(q):
        raise ValueError("w and q must have the same number of features.")
    else:
        quantile_res = np.array(
            [
                quantile_weighted_unidim(
                    np.expand_dims(
                        a.take(i, axis=feature_axis), axis=feature_axis
                    ),
                    q[i],
                    w[..., i],
                    axis=axis,
                )
                for i in range(len(q))
            ]
        )

    return np.squeeze(
        np.transpose(quantile_res, (*range(1, quantile_res.ndim), 0))
    )


def hungarian_assignment(predicted_bboxes: np.ndarray, true_bboxes: np.ndarray, min_iou:float=0.5):
    """
    Assign predicted bounding boxes to labeled ones based on maximizing IOU.
    
    This function relies on the Hungarian algorithm (also known as the 
    Kuhn-Munkres algorithm) to perform the assignment.


    :param np.ndarray predicted_bboxes: Array of predicted bounding boxes with 
        shape (N, 4), where N is the number of predictions.
    :param np.ndarray true_bboxes: Array of true bounding boxes with shape 
        (M, 4), where M is the number of true classes.
    :param float min_iou: Minimum IoU threshold to consider a prediction as 
        valid, by default 0.5.

    :return: Tuple containing:
        - Array of aligned predicted bounding boxes that have IoU greater than 
            the minimum threshold.
        - Array of true bounding boxes that correspond to the valid predicted 
            bounding boxes.
    :rtype: tuple(np.ndarray, np.ndarray)

    .. note::
        This function pads the predicted bounding boxes to match the number of 
        true bounding boxes if necessary. It then calculates the IoU matrix 
        between true and predicted bounding boxes and performs linear sum 
        assignment to maximize the total IoU. Finally, it filters out the 
        bounding boxes that do not meet the minimum IoU threshold.

    .. code-block:: python

        Examples
        --------
        >>> predicted_bboxes = np.array([[10, 10, 50, 50], [20, 20, 60, 60]])
        >>> true_bboxes = np.array([[12, 12, 48, 48], [22, 22, 58, 58], [30, 30, 70, 70]])
        >>> hungarian_assignment(predicted_bboxes, true_bboxes, min_iou=0.5)
        (array([[10, 10, 50, 50], [20, 20, 60, 60]]), array([[12, 12, 48, 48], [22, 22, 58, 58]]))
        
    """
    # Pad predicted bounding boxes to match the number of labeled ones
    def pad_predictions(predictions, labels):
        num_preds = predictions.shape[0]
        num_labels = labels.shape[0]

        if num_preds < num_labels:
            padded_predictions = np.zeros_like(labels)
            padded_predictions[:num_preds] = predictions
        else:
            padded_predictions = predictions.copy()

        return padded_predictions

    # Pad predicted bounding boxes to match the number of true bounding boxes
    padded_predictions = pad_predictions(predicted_bboxes, true_bboxes)

    # Calculate IoUs between true and predicted bounding boxes
    iou_matrix = np.round(iou(true_bboxes, padded_predictions), 2)

    # Perform linear sum assignment to maximize the total IoU
    _, best_pred_indices = linear_sum_assignment(iou_matrix, maximize=True)

    # Align predicted bounding boxes with true ones based on the best assignment
    aligned_predictions = padded_predictions[best_pred_indices]

    # Keep only those bounding boxes that have IoU greater than the minimum threshold
    valid_indices = iou(true_bboxes, aligned_predictions).diagonal() > min_iou

    return aligned_predictions[valid_indices], true_bboxes[valid_indices], valid_indices



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements conformal anomaly detection procedures.
"""
import logging
from typing import Iterable
from typing import Optional
from typing import Tuple

import numpy as np

from deel.puncc.api.calibration import ScoreCalibrator
from deel.puncc.api.splitting import IdSplitter
from deel.puncc.api.splitting import RandomSplitter

logger = logging.getLogger(__name__)


class SplitCAD:
    """Split conformal anomaly detection method based on Laxhammar's algorithm.
    The anomaly detection is based on the calibrated threshold (through
    conformal prediction) of underlying anomaly detection (model's) scores.
    For more details, we refer the user to the :ref:`theory overview
    page <theory_overview>`.

    :param BasePredictor predictor: a predictor implementing fit and predict.
    :param bool train: if False, prediction model(s) will not be (re)trained.
        Defaults to True.
    :param float random_state: random seed used when the user does not
        provide a custom fit/calibration split in `fit` method.

    Example::

        import numpy as np
        from sklearn.ensemble import IsolationForest
        from sklearn.datasets import make_moons
        import matplotlib.pyplot as plt

        from deel.puncc.anomaly_detection import SplitCAD
        from deel.puncc.api.prediction import BasePredictor

        # We generate the two moons dataset
        dataset = 4 * make_moons(n_samples=1000, noise=0.05, random_state=0)[
            0
        ] - np.array([0.5, 0.25])

        # We generate uniformly new (test) data points
        rng = np.random.RandomState(42)
        z_test = rng.uniform(low=-6, high=6, size=(150, 2))


        # The nonconformity scores are defined as the IF scores (anomaly score).
        # By default, score_samples return the opposite of IF scores.
        # We need to redefine the predict to output the nonconformity scores.
        class ADPredictor(BasePredictor):
            def predict(self, X):
                return -self.model.score_samples(X)

        # Instantiate the Isolation Forest (IF) anomaly detection model
        # and wrap it in a predictor
        if_predictor = ADPredictor(IsolationForest(random_state=42))

        # Instantiate CAD on top of IF predictor
        if_cad = SplitCAD(if_predictor, train=True, random_state=0)

        # Fit the IF on the proper fitting dataset and
        # calibrate it using calibration dataset.
        # The two datasets are sampled randomly with a ration of 7:3,
        # respectively.
        if_cad.fit(z=dataset, fit_ratio=0.7)

        # We set the maximum false detection rate to 1%
        alpha = 0.01

        # The method `predict` is called on the new data points
        # to test which are anomalous and which are not
        results = if_cad.predict(z_test, alpha=alpha)

        anomalies = z_test[results]
        not_anomalies = z_test[np.invert(results)]

        # Plot results
        plt.scatter(dataset[:, 0], dataset[:, 1], s=10, label="Inliers")
        plt.scatter(
            anomalies[:, 0],
            anomalies[:, 1],
            marker="x",
            color="red",
            s=40,
            label="Anomalies",
        )
        plt.scatter(
            not_anomalies[:, 0],
            not_anomalies[:, 1],
            marker="x",
            color="blue",
            s=40,
            label="Normal",
        )
        plt.xticks(())
        plt.yticks(())
        plt.legend()
    """

    def __init__(self, predictor, *, train=True, random_state: float = None):
        self.predictor = predictor
        self.calibrator = ScoreCalibrator(nonconf_score_func=predictor.predict)

        self.train = train

        self.random_state = random_state

        self.__is_fit = False

    def fit(
        self,
        *,
        z: Optional[Iterable] = None,
        fit_ratio: float = 0.8,
        z_fit: Optional[Iterable] = None,
        z_calib: Optional[Iterable] = None,
        **kwargs: Optional[dict],
    ):
        """This method fits the models on the fit data
        and computes nonconformity scores on calibration data.
        If z are provided, randomly split data into
        fit and calib subsets w.r.t to the fit_ratio.
        In case z_fit and z_calib are provided,
        the conformalization is performed on the given user defined
        fit and calibration sets.

        .. NOTE::

            If z is provided, `fit` ignores
            any user-defined fit/calib split.


        :param Iterable z: data points from the training dataset.
        :param float fit_ratio: the proportion of samples assigned to the
            fit subset.
        :param Iterable z_fit: data points from the fit dataset.
        :param Iterable z_calib: data points from the calibration dataset.
        :param dict kwargs: predict configuration to be passed to the model's
            fit method.

        :raises RuntimeError: no dataset provided.

        """

        if z is not None:
            splitter = RandomSplitter(
                ratio=fit_ratio, random_state=self.random_state
            )

        elif z_fit is not None and z_calib is not None:
            splitter = IdSplitter(z_fit, z_fit, z_calib, z_calib)

        elif (
            self.predictor.is_trained and z_fit is None and z_calib is not None
        ):
            splitter = IdSplitter(
                np.empty_like(z_calib), np.empty_like(z_calib), z_calib, z_calib
            )

        else:
            raise RuntimeError("No dataset provided.")

        # Apply splitter
        z_fit, _, z_calib, _ = splitter(z, z)[0]

        # Fit underlying model and calibrator
        if self.train:
            logger.info("Fitting model")
            self.predictor.fit(z_fit, **kwargs)

        # Make sure that predictor is already trained if train arg is False
        elif self.train is False and self.predictor.is_trained is False:
            raise RuntimeError(
                "'train' argument is set to 'False' but model is not pre-trained"
            )

        else:  # Skipping training
            logger.info("Skipping training.")

        # Fitting calibrator
        self.calibrator.fit(z_calib)

        self.__is_fit = True

    def predict(self, z_test: Iterable, alpha) -> Tuple[np.ndarray]:
        """Predict whether each example is an anomaly or not. The decision is
        taken based on the calibrated threshold (through conformal prediction)
        of underlying anomaly detection scores.

        :param Iterable z_test: new data points.
        :param float alpha: target maximum FDR.

        :returns: outlier tag. True if outlier, False otherwise.
        :rtype: Iterables[bool]

        """

        if self.__is_fit is None:
            raise RuntimeError("Fit method should be called before predict.")

        anomaly_pred = np.invert(
            self.calibrator.is_conformal(z_test, alpha=alpha)
        )

        return anomaly_pred



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides data splitting schemes.
"""
import importlib
from abc import ABC
from typing import Iterable
from typing import List
from typing import Tuple

import numpy as np
from sklearn import model_selection

from deel.puncc.api.utils import features_len_check
from deel.puncc.api.utils import sample_len_check
from deel.puncc.api.utils import supported_types_check

if importlib.util.find_spec("pandas") is not None:
    import pandas as pd


class BaseSplitter(ABC):
    """Abstract structure of a splitter. A splitter provides a function
    that assignes data points to fit and calibration sets.

    :param int random_state: seed to control random generation.
    """

    def __init__(self, random_state=None) -> None:
        self.random_state = random_state

    def __call__(self, *args, **kwargs) -> Tuple[np.ndarray]:
        raise NotImplementedError("Use a subclass of Splitter.")


class IdSplitter(BaseSplitter):
    """Identity splitter that wraps an already existing data assignment.

    :param Iterable X_fit: Fit features.
    :param Iterable y_fit: Fit labels.
    :param Iterable X_calib: calibration features.
    :param Iterable y_calib: calibration labels.
    """

    def __init__(
        self,
        X_fit: Iterable,
        y_fit: Iterable,
        X_calib: Iterable,
        y_calib: Iterable,
    ):
        super().__init__(random_state=None)

        # Checks
        supported_types_check(X_fit, y_fit)
        supported_types_check(X_calib, y_calib)
        sample_len_check(X_fit, y_fit)
        sample_len_check(X_calib, y_calib)
        features_len_check(X_fit, X_calib)

        self._split = [(X_fit, y_fit, X_calib, y_calib)]

    def __call__(self, X=None, y=None) -> Tuple[Iterable]:
        """Wraps into a splitter the provided fit and calibration subsets.

        :param Iterable X: features array. Not needed here, just a placeholder
            for interoperability.
        :param Iterable y: labels array. Not needed here, just a placeholder
            for interoperability.

        :returns: List of one tuple of deterministic subsets
            (X_fit, y_fit, X_calib, y_calib).
        :rtype: List[Tuple[Iterable]]
        """
        return self._split


class RandomSplitter(BaseSplitter):
    """Random splitter that assign samples given a ratio.

    :param float ratio: ratio of data assigned to the training
        (1-ratio to calibration).
    :param int random_state: seed to control random generation.
    """

    def __init__(self, ratio, random_state=None):
        if (ratio <= 0) or (ratio >= 1):
            raise ValueError(f"Ratio must be in (0,1). Provided value: {ratio}")
        self.ratio = ratio
        super().__init__(random_state=random_state)

    def __call__(
        self,
        X: Iterable,
        y: Iterable,
    ) -> Tuple[Iterable]:
        """Implements a random split strategy.

        :param Iterable X: features array.
        :param Iterable y: labels array.

        :returns: List of one tuple of random subsets
            (X_fit, y_fit, X_calib, y_calib).
        :rtype: List[Tuple[Iterable]]
        """
        # Checks
        supported_types_check(X, y)
        sample_len_check(X, y)

        rng = np.random.RandomState(seed=self.random_state)
        fit_idxs = rng.rand(len(X)) > self.ratio
        cal_idxs = np.invert(fit_idxs)
        return [(X[fit_idxs], y[fit_idxs], X[cal_idxs], y[cal_idxs])]


class KFoldSplitter(BaseSplitter):
    """KFold data splitter.

    :param int K: number of folds to generate.
    :param int random_state: seed to control random generation.
    """

    def __init__(self, K: int, random_state=None) -> None:
        if K < 2:
            raise ValueError(f"K must be >= 2. Provided value: {K}.")
        self.K = K
        super().__init__(random_state=random_state)

    def __call__(
        self,
        X: Iterable,
        y: Iterable,
    ) -> List[Tuple[Iterable]]:
        """Implements a K-fold split strategy.

        :param Iterabler X: features array.
        :param Iterable y: labels array.

        :returns: list of K split folds. Each fold is a tuple
            (X_fit, y_fit, X_calib, y_calib).
        :rtype: List[Tuple[Iterable]]
        """
        # Checks
        supported_types_check(X, y)
        sample_len_check(X, y)

        kfold = model_selection.KFold(
            self.K, shuffle=True, random_state=self.random_state
        )
        folds = []

        for fit, calib in kfold.split(X):
            if importlib.util.find_spec("pandas") is not None and isinstance(
                X, pd.DataFrame
            ):
                if isinstance(y, pd.DataFrame):
                    folds.append(
                        (
                            X.iloc[fit],
                            y.iloc[fit],
                            X.iloc[calib],
                            y.iloc[calib],
                        )
                    )
                else:
                    folds.append((X.iloc[fit], y[fit], X.iloc[calib], y[calib]))

            else:
                bool_fit_idx = np.array([i in fit for i in range(len(X))])
                folds.append(
                    (
                        X[bool_fit_idx],
                        y[bool_fit_idx],
                        X[~bool_fit_idx],
                        y[~bool_fit_idx],
                    )
                )

        return folds



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides the canvas for conformal prediction.
"""
import logging
import pickle
from copy import deepcopy
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from deel.puncc.api.calibration import BaseCalibrator
from deel.puncc.api.calibration import CvPlusCalibrator
from deel.puncc.api.corrections import bonferroni
from deel.puncc.api.prediction import BasePredictor
from deel.puncc.api.prediction import DualPredictor
from deel.puncc.api.splitting import BaseSplitter
from deel.puncc.api.splitting import IdSplitter

logger = logging.getLogger(__name__)


class ConformalPredictor:
    """Conformal predictor class.

    :param deel.puncc.api.prediction.BasePredictor | object predictor:
        underlying model to be conformalized. The model can directly be
        passed as argument if it already has `fit` and `predict` methods.
    :param deel.puncc.api.prediction.BaseCalibrator calibrator: nonconformity
        computation strategy and set predictor.
    :param deel.puncc.api.prediction.BaseSplitter splitter: fit/calibration
        split strategy. The splitter can be set to None if the underlying
        model is pretrained.
    :param str method: method to handle the ensemble prediction and calibration
        in case the splitter is a K-fold-like strategy. Defaults to 'cv+' to
        follow cv+ procedure.
    :param bool train: if False, prediction model(s) will not be (re)trained.
        Defaults to True.

    .. WARNING::
        if a K-Fold-like splitter is provided with the :data:`train` attribute
        set to True, an exception is raised.
        The models have to be trained during the call :meth:`fit`.


    **Conformal Regression example:**

    .. code-block:: python

        from sklearn import linear_model
        from sklearn.datasets import make_regression
        from sklearn.model_selection import train_test_split

        from deel.puncc.api.conformalization import ConformalPredictor
        from deel.puncc.api.prediction import BasePredictor
        from deel.puncc.api.calibration import BaseCalibrator
        from deel.puncc.api.splitting import KFoldSplitter
        from deel.puncc.api import nonconformity_scores
        from deel.puncc.api import prediction_sets

        # Generate a random regression problem
        X, y = make_regression(n_samples=1000, n_features=4, n_informative=2,
                                random_state=0, shuffle=False)

        # Split data into train and test
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=.2, random_state=0
        )

        # Regression linear model
        model = linear_model.LinearRegression()

        # Definition of a predictor. Note that it is not required to wrap
        # the model here because it already implements fit and predict methods
        predictor = BasePredictor(model)

        # Definition of a calibrator, built for a given nonconformity scores
        # and a procedure to build the prediction sets
        calibrator = BaseCalibrator(nonconf_score_func=nonconformity_scores.mad,
                                pred_set_func=prediction_sets.constant_interval)

        # Definition of a K-fold splitter that produces
        # 20 folds of fit/calibration
        kfold_splitter = KFoldSplitter(K=20, random_state=42)

        # Conformal predictor requires the three components instantiated
        # previously. Our choice of calibrator and splitter yields a cv+ procedure
        conformal_predictor = ConformalPredictor(predictor=predictor,
                                                calibrator=calibrator,
                                                splitter=kfold_splitter,
                                                train=True)

        # Fit model and compute nonconformity scores
        conformal_predictor.fit(X_train, y_train)

        # The lower and upper bounds of the prediction interval are predicted
        # by the call to predict on the new data w.r.t a risk level of 10%.
        # Besides, there is no aggregate point prediction in cv+ so y_pred is None.
        y_pred , y_lower, y_upper = conformal_predictor.predict(X_test, alpha=.1)

    """

    def __init__(
        self,
        calibrator: BaseCalibrator,
        predictor: Union[BasePredictor, object],
        splitter: BaseSplitter,
        method: str = "cv+",
        train: bool = True,
    ):
        self.calibrator = calibrator

        if isinstance(predictor, (BasePredictor, DualPredictor)):
            self.predictor = predictor

        elif not hasattr(predictor, "predict"):
            raise RuntimeError(
                "Provided model has no predict method. "
                + "Use a BasePredictor or a DualPredictor to build "
                + "a compatible predictor."
            )

        elif train and not hasattr(predictor, "fit"):
            raise RuntimeError(
                "Provided model is not trained and has no fit method. "
                + "Use a BasePredictor or a DualPredictor to build "
                + "a compatible predictor."
            )

        else:
            self.predictor = BasePredictor(predictor, is_trained=not train)

        if train and splitter is None:
            raise RuntimeError(
                "The splitter argument is None but train is set to True. "
                + "Please provide a correct splitter to train the underlying "
                + "model."
            )

        if method != "cv+":
            raise RuntimeError(
                f"Method {method} is not implemented." + "Please choose 'cv+'."
            )

        self.splitter = splitter
        self.method = method
        self.train = train
        self._cv_cp_agg = None

    def get_nonconformity_scores(self) -> dict:
        """Getter for computed nonconformity scores on the calibration(s) set(s).

        :returns: dictionary of nonconformity scores indexed by the fold index.
        :rtype: dict

        :raises RuntimeError: :meth:`fit` needs to be called before
            :meth:`get_nonconformity_scores`.
        """

        if self._cv_cp_agg is None:
            raise RuntimeError("Error: call 'fit' method first.")

        return self._cv_cp_agg.get_nonconformity_scores()

    def get_weights(self) -> dict:
        """Getter for weights associated to calibration samples.

        :returns: dictionary of weights indexed by the fold index.
        :rtype: dict

        :raises RuntimeError: :meth:`fit` needs to be called before
            :meth:`get_weights`.
        """

        if self._cv_cp_agg is None:
            raise RuntimeError("Error: call 'fit' method first.")

        return self._cv_cp_agg.get_weights()

    def fit(
        self,
        X: Iterable,
        y: Iterable,
        use_cached: bool = False,
        **kwargs,
    ) -> None:
        """Fit the model(s) and estimate the nonconformity scores.

        If the splitter is an instance of
        :class:`deel.puncc.splitting.KFoldSplitter`, the fit operates on each
        fold separately. Thereafter, the predictions and nonconformity scores
        are combined accordingly to an aggregation method (cv+ by default).

        :param Iterable X: features.
        :param Iterable y: labels.
        :param bool use_cached: if set, enables to add the previously computed
            nonconformity scores (if any) to the pool estimated in the current
            call to `fit`. The aggregation follows the CV+
            procedure.
        :param dict kwargs: options configuration for the training.

        :raises RuntimeError: inconsistencies between the train status of the
            model(s) and the :data:`train` class attribute.
        """
        # Get split folds. Each fold split is a iterable of a quadruple that
        # contains fit and calibration data.
        if self.splitter is None:
            splits = IdSplitter(X, y, X, y)()
        else:
            splits = self.splitter(X, y)

        # The Cross validation aggregator will aggregate the predictors and
        # calibrators fitted on each of the K splits.
        if self._cv_cp_agg is None or use_cached is False:
            cached_len = 0
            self._cv_cp_agg = CrossValCpAggregator(
                K=len(splits), method=self.method
            )
        else:
            cached_len = self._cv_cp_agg.K
            self._cv_cp_agg.K = cached_len + len(splits)

        # In case of multiple split folds, the predictor require training.
        # Having 'self.train' set to False is therefore an inconsistency
        if len(splits) > 1 and not self.train:
            raise RuntimeError(
                "Model already trained. This is inconsistent with the"
                + "cross-validation strategy."
            )

        # Core loop: for each split (that contains fit and calib data):
        #   1- The predictor f_i is fitted of (X_fit, y_fit) (if necessary)
        #   2- y_pred is predicted by f_i
        #   3- The calibrator is fitted to approximate the distribution of
        #      nonconformity scores
        for i, (X_fit, y_fit, X_calib, y_calib) in enumerate(splits):
            # Make local copies of the structure of the predictor and the calibrator.
            # In case of a K-fold like splitting strategy, these structures are
            # inherited by the predictor/calibrator used in each fold.
            if len(splits) > 1:
                predictor = self.predictor.copy()
                calibrator = deepcopy(self.calibrator)
            else:
                predictor = self.predictor
                calibrator = self.calibrator

            if self.train:
                if self.splitter is None:
                    raise RuntimeError(
                        "The splitter argument is None but train is set to "
                        + "True. Please provide a correct splitter to train "
                        + "the underlying model."
                    )
                logger.info(f"Fitting model on fold {i+cached_len}")
                predictor.fit(X_fit, y_fit, **kwargs)  # Fit K-fold predictor

            # Make sure that predictor is already trained if train arg is False
            elif self.train is False and predictor.get_is_trained() is False:
                raise RuntimeError(
                    "'train' argument is set to 'False' but model(s) not pre-trained."
                )

            else:  # Skipping training
                logger.info("Skipping training.")

            # Call predictor to estimate predictions
            logger.info(f"Model predictions on X_calib fold {i+cached_len}")
            y_pred = predictor.predict(X_calib)
            logger.debug("Shape of y_pred")

            # Fit calibrator
            logger.info(f"Fitting calibrator on fold {i+cached_len}")
            calibrator.fit(y_true=y_calib, y_pred=y_pred)

            # Compute normalized weights of the nonconformity scores
            # if a weight function is provided
            if calibrator.weight_func:
                weights = calibrator.weight_func(X_calib)
                norm_weights = calibrator.barber_weights(weights=weights)
                # Store the mornalized weights
                calibrator.set_norm_weights(norm_weights)

            # Add predictor and calibrator to the collection that is used later
            # by the predict method
            self._cv_cp_agg.append_predictor(i + cached_len, predictor)
            self._cv_cp_agg.append_calibrator(i + cached_len, calibrator)

    def predict(
        self,
        X: Iterable,
        alpha: float,
        correction_func: Optional[Callable] = bonferroni,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]:
        """Predict point, and interval estimates for X data.

        :param Iterable X: features.
        :param float alpha: significance level (max miscoverage target).
        :param Callable correction_func: correction for multiple hypothesis
            testing in the case of multivariate regression. Defaults to
            Bonferroni correction.

        :returns: (y_pred, y_lower, y_higher) or (y_pred, pred_set).
        :rtype: Union[Tuple[np.ndarray, np.ndarray],
            Tuple[np.ndarray, np.ndarray, np.ndarray]]
        """
        if self._cv_cp_agg is None:
            raise RuntimeError("Error: call 'fit' method first.")

        return self._cv_cp_agg.predict(X, alpha, correction_func)

    def save(self, path, save_data=True):
        """Serialize current conformal predictor and write it to a file.

        :param str path: File path.
        
        :param bool save_data: If True, save the custom data used to 
            fit/calibrate the model.
        
        """
        # Remove cached data if needed (case of IdSplitter)
        is_cached = False
        if save_data and hasattr(self.splitter, "_split"):
            cached = self.splitter._split
            is_cached = True
            self.splitter._split = None
            print("\033[33m\033[1mWarning:\033[0m Custom train/calibration data removed from the"
                " conformal predictor. If you want to keep them,"
                " please set flag `save_data` to True.")

        with open(path, "wb") as output_file:
            pickle.dump(self.__dict__, output_file)
        if is_cached:
            self.splitter._split = cached

    @staticmethod
    def load(path):
        """Load conformal predictor from a file.

        :param str path: file path.

        :returns: loaded conformal predictor instance.
        :rtype: ConformalPredictor
        """
        with open(path, "rb") as input_file:
            saved_dict = pickle.load(input_file)

        loaded_cp = ConformalPredictor(
            calibrator=None, predictor=BasePredictor(None), splitter=object()
        )
        loaded_cp.__dict__.update(saved_dict)
        return loaded_cp


class CrossValCpAggregator:
    """This class enables to aggregate predictions and calibrations
    from different K-folds.


    :param int K: number of folds
    :param dict _predictors: collection of predictors fitted on the K-folds
    :param dict _calibrators: collection of calibrators fitted on the K-folds
    :param str method: method to handle the ensemble prediction and
        calibration, defaults to 'cv+'.
    """

    def __init__(
        self,
        K: int,
        method: str = "cv+",
    ):
        self.K = K  # Number of K-folds
        self._predictors = {}
        self._calibrators = {}

        if method not in ("cv+"):
            raise NotImplementedError(
                f"Method {method} is not implemented. " + "Please choose 'cv+'."
            )

        self.method = method

    def append_predictor(self, key, predictor):
        """Add predictor in kfold predictors dictionnary.

        :param int key: key of the predictor.
        :param BasePredictor|DualPredictor predictor: predictor to be appended.

        """
        self._predictors[key] = predictor.copy()

    def append_calibrator(self, key, calibrator):
        """Add calibrator in kfold calibrators dictionnary.

        :param int key: key of the calibrator.
        :param BaseCalibrator predictor: calibrator to be appended.

        """
        self._calibrators[key] = deepcopy(calibrator)

    def get_nonconformity_scores(self) -> dict:
        """Get a dictionnary of residuals computed on the K-folds.

        :returns: dictionary of residual indexed by the K-fold number.
        :rtype: dict
        """
        return {
            k: calibrator.get_nonconformity_scores()
            for k, calibrator in self._calibrators.items()
        }

    def get_weights(self) -> dict:
        """Get a dictionnary of normalized weights computed on the K-folds.

        :returns: dictionary of normalized weights indexed by the K-fold number.
        :rtype: dict
        """
        return {
            k: calibrator.get_weights()
            for k, calibrator in self._calibrators.items()
        }

    def predict(
        self,
        X: Iterable,
        alpha: float,
        correction_func: Optional[Callable] = bonferroni,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]:  #  type: ignore
        """Predict point, interval and variability estimates for X data.

        :param Iterable X: features.
        :param float alpha: significance level (max miscoverage target).
        :param Callable correction_func: correction for multiple hypothesis
            testing in the case of multivariate regression. Defaults to
            Bonferroni correction.

        :returns: y_pred, y_lower, y_higher.
        :rtype: Union[Tuple[np.ndarray, np.ndarray],
            Tuple[np.ndarray, np.ndarray, np.ndarray]]
        """
        assert (
            self._predictors.keys() == self._calibrators.keys()
        ), "K-fold predictors are not well calibrated."

        K = len(self._predictors.keys())  # Number of folds

        # No cross-val strategy if K = 1
        if K == 1:
            for k in self._predictors.keys():
                predictor = self._predictors[k]
                calibrator = self._calibrators[k]
                # Get normalized weights of the nonconformity scores
                norm_weights = calibrator.get_norm_weights()
                y_pred = predictor.predict(X=X)
                set_pred = calibrator.calibrate(
                    alpha=alpha,
                    y_pred=y_pred,
                    weights=norm_weights,
                    correction=correction_func,
                )
                return (y_pred, *set_pred)

        else:  # TODO support multireg with CV
            y_pred = None

            if self.method == "cv+":
                cvp_calibrator = CvPlusCalibrator(self._calibrators)
                set_pred = cvp_calibrator.calibrate(
                    X=X,
                    kfold_predictors_dict=self._predictors,
                    alpha=alpha,
                )
                return (y_pred, *set_pred)

            raise RuntimeError(
                f"Method {self.method} is not implemented."
                + "Please choose 'cv+'."
            )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides correction functions for multiple hypothesis testing. To be used
when building a :ref:`calibrator <calibration>` for multivariate regression or object detection.
"""
from typing import Union

import numpy as np


def bonferroni(alpha: float, nvars: int = 1) -> Union[float, np.ndarray]:
    """Bonferroni correction for multiple comparisons.

    :param float alpha: nominal coverage level.
    :param int nvars: number of output features.


    :returns: corrected coverage level.
    :rtype: float or ndarray.
    """
    # Sanity checks
    if np.any(alpha <= 0) or np.any(alpha >= 1):
        raise ValueError("alpha must be in (0,1)")

    if nvars <= 0:
        raise ValueError("nvars must be a positive integer")

    if nvars == 1:
        return alpha

    return np.ones(nvars) * alpha / nvars


def weighted_bonferroni(alpha: float, weights: np.ndarray) -> np.ndarray:
    """Weighted Bonferroni correction for multiple comparisons.

    :param float alpha: nominal coverage level.
    :param np.ndarray weights: weights associated to each output feature.


    :returns: array of corrected featurewise coverage levels.
    :rtype: np.ndarray.
    """
    # Sanity checks
    if alpha <= 0 or alpha >= 1:
        raise ValueError("alpha must be in (0,1)")

    # Positivity check
    positiveness_condition = np.all(weights > 0)
    if not positiveness_condition:
        raise RuntimeError("weights must be positive")

    # Normalization check
    norm_condition = np.isclose(np.sum(weights) - 1, 0, atol=1e-14)
    if not norm_condition:
        error = (
            "weights are not normalized. Sum of weights is"
            + f"{np.sum(weights)}"
        )
        raise RuntimeError(error)

    return alpha * weights



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Initialization of puncc."""
import logging.config


# Create the Logger
logging.basicConfig(
    format="%(asctime)s === %(name)s [%(funcName)s()] | %(levelname)s | - %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.ERROR,
)
loggers = logging.getLogger(__name__)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module provides conformal prediction metrics.
"""
from typing import Tuple

import numpy as np

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Classification ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def classification_mean_coverage(
    y_true: np.ndarray, set_pred: Tuple[np.ndarray]
) -> float:
    """Compute empirical coverage of the prediction sets.

    Given the :math:`i`-th prediction set :math:`S(X_i)`, the coverage is:

    * :math:`cov(X_i) = 1` if :math:`y_{true} \in S(X_i)`
    * :math:`cov(X_i) = 0` otherwise

    With N the number of examples, the average coverage is
    :math:`1/N \sum_{i=1}^{N} cov(X_i)`.

    :param np.ndarray y_true: Observed label
    :param Tuple[np.ndarray] set_pred: label prediction set

    :returns: average coverage, indicating the proportion of instances that are
        correctly covered.
    :rtype: float
    """
    counter = 0
    for y, S in zip(y_true, set_pred):
        if (S != []) and (y in S):
            counter += 1
    return counter / len(y_true)


def classification_mean_size(set_pred: Tuple[np.ndarray]) -> float:
    """Compute average size of the prediction sets.

    :param Tuple[np.ndarray] set_pred: label prediction set

    :returns: Average size of the prediction sets
    :rtype: float
    """
    return np.mean([len(s) for s in set_pred])


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Regression ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def regression_mean_coverage(y_true, y_pred_lower, y_pred_upper) -> float:
    """Compute average coverage on several prediction intervals.

    Given the :math:`i`-th prediction interval :math:`C(X_i)`, the coverage is:

        * :math:`cov(X_i) = 1` if :math:`y_{true} \in C(X_i)`
        * :math:`cov(X_i) = 0` otherwise

    With N the number of examples, the average coverage is
    :math:`1/N \sum_{i=1}^{N} cov(X_i)`.

    :param ndarray y_true: label true values.
    :param ndarray y_pred_lower: lower bounds of the prediction intervals.
    :param ndarray y_pred_upper: upper bounds of the prediction intervals.

    :returns: average coverage, indicating the proportion of instances that are
        correctly covered.
    :rtype: float
    """
    return ((y_true >= y_pred_lower) & (y_true <= y_pred_upper)).mean()


def regression_ace(y_true, y_pred_lower, y_pred_upper, alpha) -> float:
    """Compte the Average Coverage Error (ACE).

    :param ndarray y_true: label true values.
    :param ndarray y_pred_lower: lower bounds of the prediction intervals.
    :param ndarray y_pred_upper: upper bounds of the prediction intervals.
    :param float alpha: significance level (max miscoverage target).

    .. NOTE::

        The ACE is the distance between the nominal coverage :math:`1-\\alpha`
        and the empirical average coverage :math:`AC` such that
        :math:`ACE = AC - (1-\\alpha)`.

        If the ACE is strictly negative, the prediction intervals are
        marginally undercovering. If the ACE is strictly positive, the
        prediction intervals are maginally conservative.

    :returns: the average coverage error (ACE).
    :rtype: float
    """
    cov = regression_mean_coverage(y_true, y_pred_lower, y_pred_upper)
    return cov - (1 - alpha)


def regression_sharpness(y_pred_lower, y_pred_upper) -> float:
    """Compute the average absolute width of the prediction intervals.

    :param ndarray y_pred_lower: lower bounds of the prediction intervals.
    :param ndarray y_pred_upper: upper bounds of the prediction intervals.

    :returns: average absolute width of the prediction intervals.
    :rtype: float
    """
    return (np.abs(y_pred_upper - y_pred_lower)).mean()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Object Detection ~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def object_detection_mean_coverage(
    y_pred_outer: np.ndarray, y_true: np.ndarray
):
    """
    Calculate the mean coverage of conformal object detection predictions.
    For each instance, coverage is defined as the true bounding box being inside
    the predicted outer bounding box.

    :param np.ndarray y_pred: array of predicted outer bounding boxes with shape (n, 4).
    :type y_pred: np.ndarray
    :param np.ndarray y_true: array of true bounding boxes with shape (n, 4).

    :return: average coverage, indicating the proportion of objects that are
        correctly covered.
    :rtype: float

    """
    x_min_true, y_min_true, x_max_true, y_max_true = np.hsplit(y_true, 4)
    x_min, y_min, x_max, y_max = np.hsplit(y_pred_outer, 4)
    cov = (
        (x_min <= x_min_true)
        * (y_min <= y_min_true)
        * (x_max >= x_max_true)
        * (y_max >= y_max_true)
    )
    return np.mean(cov)


def object_detection_mean_area(y_pred: np.ndarray):
    """
    Calculate the mean area of object bounding predictions.

    :param np.ndarray y_pred: array of predicted bounding boxes with shape (n, 4).

    :return: average area of the bounding boxes
    :rtype: float

    """
    x_min, y_min, x_max, y_max = np.hsplit(y_pred, 4)
    return np.mean((x_max - x_min) * (y_max - y_min))


# Calculate Intersection over Union (IOU) between two bounding boxes
def iou(bboxes1: np.ndarray, bboxes2: np.ndarray) -> np.ndarray:
    """
    Calculates the Intersection over Union (IoU) between two sets of 
    bounding boxes. The IoU is calculated as the ratio between the area of 
    intersection and the area of union between two bounding boxes.
    
    :param np.ndarray bboxes1: array of shape (N, 4) representing the 
        coordinates of N bounding boxes in the format 
        [x_min, y_min, x_max, y_max].
    :type y_pred: np.ndarray
    :param np.ndarray bboxes2: array of shape (N, 4) representing the 
        coordinates of N bounding boxes in the format 
        [x_min, y_min, x_max, y_max].

    :return: iou (numpy.ndarray): Array of shape (N, ) representing the IoU 
        between each pair of bounding boxes.
    :rtype: np.ndarray

    """

    x1_min, y1_min, x1_max, y1_max = np.split(bboxes1, 4, axis=1)
    x2_min, y2_min, x2_max, y2_max = np.split(bboxes2, 4, axis=1)

    inter_x_min = np.maximum(x1_min, np.transpose(x2_min))
    inter_y_min = np.maximum(y1_min, np.transpose(y2_min))
    inter_x_max = np.minimum(x1_max, np.transpose(x2_max))
    inter_y_max = np.minimum(y1_max, np.transpose(y2_max))

    inter_width = np.maximum(inter_x_max - inter_x_min + 1, 0)
    inter_height = np.maximum(inter_y_max - inter_y_min + 1, 0)
    inter_area = inter_width * inter_height

    box1_area = (x1_max - x1_min + 1) * (y1_max - y1_min + 1)
    box2_area = (x2_max - x2_min + 1) * (y2_max - y2_min + 1)

    result = inter_area / (box1_area + np.transpose(box2_area) - inter_area)
    return result



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements conformal object detection procedures.
"""
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from deel.puncc.api import nonconformity_scores
from deel.puncc.api import prediction_sets
from deel.puncc.api.calibration import BaseCalibrator
from deel.puncc.api.conformalization import ConformalPredictor
from deel.puncc.api.corrections import bonferroni
from deel.puncc.api.prediction import BasePredictor
from deel.puncc.api.splitting import IdSplitter
from deel.puncc.api.splitting import RandomSplitter


class SplitBoxWise:
    """Implementation of box-wise conformal object detection. For more info,
    we refer the user to the :ref:`theory overview page <theory splitboxwise>`

    :param BasePredictor | Any predictor: a predictive model.
    :param bool train: if False, prediction model(s) will not be (re)trained.
        Defaults to False.
    :param callable weight_func: function that takes as argument an array of
        features X and returns associated "conformality" weights, defaults to
        None.
    :param str method: chose between "additive" or "multiplicative" box-wise
        conformalization.
    :param int random_state: random seed used when the user does not
        provide a custom fit/calibration split in `fit` method.

    :raises ValueError: if method is not 'additive' or 'multiplicative'.

    .. _example splitboxwise:

    Example::

        from deel.puncc.object_detection import SplitBoxWise
        import numpy as np

        from sklearn.datasets import make_regression
        from sklearn.model_selection import train_test_split
        from sklearn.ensemble import RandomForestRegressor

        from deel.puncc.metrics import object_detection_mean_coverage
        from deel.puncc.metrics import object_detection_mean_area

        # Generate a random regression problem
        X, y = make_regression(
            n_samples=1000,
            n_features=4,
            n_informative=2,
            n_targets=4,
            random_state=0,
            shuffle=False,
        )

        # Create dummy object localization data formated as (x1, y1, x2, y2)
        y = np.abs(y)
        x1 = np.min(y[:, :2], axis=1)
        y1 = np.min(y[:, 2:], axis=1)
        x2 = np.max(y[:, :2], axis=1)
        y2 = np.max(y[:, 2:], axis=1)
        y = np.array([x1, y1, x2, y2]).T


        # Split data into train and test
        X, X_test, y, y_test = train_test_split(X, y, test_size=0.2, random_state=0)

        # Split train data into fit and calibration
        X_fit, X_calib, y_fit, y_calib = train_test_split(
            X, y, test_size=0.2, random_state=0
        )

        # Create a random forest model
        rf_model = RandomForestRegressor(n_estimators=100, random_state=0)

        # CP method initialization
        od_cp = SplitBoxWise(rf_model, method="multiplicative", train=True)

        # The call to `fit` trains the model and computes the nonconformity
        # scores on the calibration set
        od_cp.fit(X_fit=X_fit, y_fit=y_fit, X_calib=X_calib, y_calib=y_calib)

        # The predict method infers prediction intervals with respect to
        # the significance level alpha = 20%
        y_pred, box_inner, box_outer = od_cp.predict(X_test, alpha=0.2)

        # Compute marginal coverage and average width of the prediction intervals
        coverage = object_detection_mean_coverage(box_outer, y_test)
        average_area = object_detection_mean_area(box_outer)
        print(f"Marginal coverage: {np.round(coverage, 2)}")
    """

    def __init__(
        self,
        predictor: Union[BasePredictor, Any],
        *,
        train: bool = False,
        weight_func: Optional[Callable] = None,
        method: str = "additive",
        random_state: int = 0,
    ):
        self.predictor = predictor
        if method == "additive":
            nonconf_score_func = nonconformity_scores.difference
            pred_set_func = prediction_sets.constant_bbox
        elif method == "multiplicative":
            nonconf_score_func = nonconformity_scores.scaled_bbox_difference
            pred_set_func = prediction_sets.scaled_bbox
        else:
            raise ValueError("method must be 'additive' or 'multiplicative'.")

        self.calibrator = BaseCalibrator(
            nonconf_score_func=nonconf_score_func,
            pred_set_func=pred_set_func,
            weight_func=weight_func,
        )

        self.train = train

        self.random_state = random_state

        self.conformal_predictor = ConformalPredictor(
            predictor=self.predictor,
            calibrator=self.calibrator,
            splitter=object(),
            train=train,
        )

    def fit(
        self,
        *,
        X: Optional[Iterable] = None,
        y: Optional[Iterable] = None,
        fit_ratio: float = 0.8,
        X_fit: Optional[Iterable] = None,
        y_fit: Optional[Iterable] = None,
        X_calib: Optional[Iterable] = None,
        y_calib: Optional[Iterable] = None,
        use_cached: bool = False,
        **kwargs: Optional[dict],
    ):
        """This method fits the models on the fit data
        and computes nonconformity scores on calibration data.
        If (X, y) are provided, randomly split data into
        fit and calib subsets w.r.t to the fit_ratio.
        In case (X_fit, y_fit) and (X_calib, y_calib) are provided,
        the conformalization is performed on the given user defined
        fit and calibration sets.

        .. NOTE::

            If X and y are provided, `fit` ignores
            any user-defined fit/calib split.


        :param Iterable X: features from the training dataset.
        :param Iterable y: labels from the training dataset.
        :param float fit_ratio: the proportion of samples assigned to the
            fit subset.
        :param Iterable X_fit: features from the fit dataset.
        :param Iterable y_fit: labels from the fit dataset.
        :param Iterable X_calib: features from the calibration dataset.
        :param Iterable y_calib: labels from the calibration dataset.
        :param bool use_cached: if set, enables to add the previously computed
            nonconformity scores (if any) to the pool estimated in the current
            call to `fit`. The aggregation follows the CV+
            procedure.
        :param dict kwargs: predict configuration to be passed to the model's
            fit method.

        :raises RuntimeError: no dataset provided.

        """

        # Check if predictor is trained. Suppose that it is trained if the
        # predictor has not "is_trained" attribute
        is_trained = not hasattr(self.predictor, "is_trained") or (
            hasattr(self.predictor, "is_trained") and self.predictor.is_trained
        )

        if X is not None and y is not None:
            splitter = RandomSplitter(
                ratio=fit_ratio, random_state=self.random_state
            )

        elif (
            X_fit is not None
            and y_fit is not None
            and X_calib is not None
            and y_calib is not None
        ):
            splitter = IdSplitter(X_fit, y_fit, X_calib, y_calib)

        elif (
            is_trained
            and X_fit is None
            and y_fit is None
            and X_calib is not None
            and y_calib is not None
        ):
            if self.train is True:
                raise RuntimeError(
                    "Argument 'train' is True but no training dataset provided."
                )
            splitter = IdSplitter(
                np.empty_like(X_calib),
                np.empty_like(y_calib),
                X_calib,
                y_calib,
            )

        else:
            raise RuntimeError("No dataset provided.")

        # Update splitter
        self.conformal_predictor.splitter = splitter

        self.conformal_predictor.fit(X=X, y=y, use_cached=use_cached, **kwargs)

    def predict(
        self,
        X_test: Iterable,
        alpha: float,
        correction_func: Callable = lambda x: bonferroni(x, 4),
    ) -> Tuple[np.ndarray]:
        """Conformal object detection (w.r.t target miscoverage alpha) for
        new samples.

        :param Iterable X_test: features of new samples.
        :param float alpha: target maximum miscoverage.
        :param Callable correction_func: correction for multiple hypothesis
            testing in the case of multivariate regression. Defaults to
            Bonferroni correction.

        :returns: y_pred, y_lower, y_higher
        :rtype: Tuple[ndarray]

        """

        # Return format: y_pred, y_lower, y_higher
        return self.conformal_predictor.predict(X_test, alpha, correction_func)

    def get_nonconformity_scores(self) -> np.ndarray:
        """Get computed nonconformity scores.

        :returns: computed nonconfomity scores.
        :rtype: ndarray

        """

        # Get all nconf scores on the fitting kfolds
        kfold_nconf_scores = self.conformal_predictor.get_nonconformity_scores()

        # With split CP, the nconf scores dict has only one element if no
        # cache was used.
        if len(kfold_nconf_scores) == 1:
            return list(kfold_nconf_scores.values())[0]

        return kfold_nconf_scores
